# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
import subprocess
import sys
import cli.utils.constants as constants

from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import get_node_type

lg = nsx_logging.getLogger("cli.utils.command_utils")

_SUDO = "/usr/bin/sudo"
FALLBACK_OUTPUT_SIZE = 20971520  # 20MB


# ----------------------------------------------------------------------
# get full path for esx command
# ----------------------------------------------------------------------
def get_esx_command_path(cmd):
    """Get full path of an ESXi command.

    For ESXi, instead of Popen we are using runcommand module, which
    does not work if full command path is not specified. This method
    helps in finding location of primary executable and return its
    full path for execution via runcommand.
    """
    if os.path.isfile(cmd):
        return cmd
    else:
        path = os.environ.get('PATH', '')
        for p in path.split(os.path.pathsep):
            proposed_path = os.path.join(p, cmd)
            if os.path.isfile(proposed_path):
                return proposed_path
    raise IOError("Unable to find full path of executable: %s" % cmd)



# ----------------------------------------------------------------------
# invoke command
# ----------------------------------------------------------------------
def invoke_command(cmd_list, display=False, max_out_size=None,
                   encoding="utf-8", error="strict"):
    """Invoke a command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["ps", "aux"]).
        display: output stdout and stderr on the fly for watching provider.
        max_out_size: maximum output size in bytes, only applicable for ESX.
        encoding: Default encoding is 'utf-8'.
        error: Errors may be given to set a different error, handling scheme.
               Default is 'strict' meaning that encoding errors raise a UnicodeEncodeError.
               Other possible values are 'ignore', 'replace' and 'xmlcharrefreplace'
               as well as any other name registered with codecs.register_error
               that can handle UnicodeEncodeErrors.
    Return:
        if display = True: (returncode, None, None) # Not supported on ESX
        else: (returncode, output, error)
    """
    lg.debug("Invoking: %s" % cmd_list)

    if get_node_type() == NODE_TYPE_HOST_ESX:
        # display = True is not supported by vmware.runcommand module
        if display == True:
            return (1, "", "Displaying to stdout/stderr is not supported")
        # On ESX, use a simpler version of Popen (called by runcommand)
        # Related PR https://bugzilla.eng.vmware.com/show_bug.cgi?id=2007499
        from vmware.runcommand import runcommand, RunCommandError,\
                                      MAXOUTPUTSIZE
        cmd_list[0] = get_esx_command_path(cmd_list[0])
        str_all = lambda iterable: [str(i) for i in iterable]

        if max_out_size is None:
            try:
                rc, out, err = runcommand(cmd_list, redirectErr=False)
            except RunCommandError as ex:
                # Incase of failure with default permissible output, retry with fallback size
                # Related PR https://bugzilla.eng.vmware.com/show_bug.cgi?id=2388272
                if "large" in " ".join(str_all(ex.args)):
                    lg.warn("Output for command '%s' exceeds %s bytes, "\
                             "retrying with %s bytes" % (cmd_list, MAXOUTPUTSIZE,
                                                        FALLBACK_OUTPUT_SIZE))
                    rc, out, err = runcommand(cmd_list, redirectErr=False,
                                              maxOutputSize=FALLBACK_OUTPUT_SIZE)
                else:
                    raise ex.__class__(ex.args)
        else:
            rc, out, err = runcommand(cmd_list, redirectErr=False,
                                      maxOutputSize=max_out_size)
        if hasattr(out, "decode"):
            out = out.decode(encoding, error)
        if hasattr(err, "decode"):
            err = err.decode(encoding, error)
        return (rc, out, err)

    if display:
        out_stream = sys.stdout
        err_stream = sys.stdout
    else:
        out_stream = subprocess.PIPE
        err_stream = subprocess.PIPE
    p = subprocess.Popen(cmd_list,
                         stdout=out_stream,
                         stderr=err_stream)
    if display:
        p.wait()
    out, err = p.communicate()
    if sys.version_info >= (3,):
        # In Python3, communicate() returns byte strings, so convert to string
        if out is not None:
            out = out.decode(encoding, error)
        if err is not None:
            err = err.decode(encoding, error)
    rc = p.returncode
    lg.debug("rc: %d\nerr: %s\nout: %s\n", rc, err, out)
    return rc, out, err


# ----------------------------------------------------------------------
# Follow a file until Ctrl+C is pressed.
# ----------------------------------------------------------------------
def tail_follow(logfile, sudo=False, filter=None):
    """Open and follow the given log file, optionally with a filter.

    NOTE: This method is setup for following log files under /var/log/...
    and the sudoers file has been configured as such. If sudo is not required
    then any text file can be followed.

    Use Ctrl+C to stop following the file.

    Args:
        logfile: The log file to follow.
        sudo: Boolean. If True, prepend the tail command with sudo.
        filter: Only show lines from the file that match the filter (uses grep).

    Return:
        True when Ctrl+C is used to stop following the file.
        False if the tail command failed to start.
    """
    if not os.path.exists(logfile):
        return False
    result = True
    filtercmd = None
    tailcmd = ['/usr/bin/tail', '-F', logfile]
    if sudo:
        # Validate that the log file is under /var/log when sudo is requested.
        if not logfile.startswith('/var/log/'):
            lg.error("The log file must be under /var/log/",
                     error_code=ec.INTERNAL_ERROR)
            return False
        tailcmd.insert(0, _SUDO)
    if filter:
        filtercmd = ['grep', filter]
        cmd_stdout = subprocess.PIPE
    tail_po = filter_po = None
    try:
        cmd_stdout = subprocess.PIPE if filtercmd else None
        tail_po = subprocess.Popen(tailcmd, stdout=cmd_stdout)
        if filtercmd:
            filter_po = subprocess.Popen(filtercmd, stdin=tail_po.stdout)
    except KeyboardInterrupt:
        pass
    except Exception:
        result = False
        msg = "Error invoking: %s" % " ".join(tailcmd)
        if filtercmd:
            msg = "%s | %s" % (msg, " ".join(filtercmd))
        lg.exception(msg, error_code=ec.INTERNAL_ERROR)
    finally:
        if filter_po:
            err = None
            try:
                _, err = filter_po.communicate()
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            except Exception as e:
                result = False
                err = e
            if err:
                lg.error("Error invoking: grep, err: %s", err,
                         error_code=ec.OPERATION_FAILED)
        if tail_po:
            err = None
            try:
                _, err = tail_po.communicate()
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            except ValueError:  # I/O operation on closed file
                pass
            except Exception as e:
                result = False
                err = e
            if err:
                lg.error("Error invoking: %s, err: %s",
                         " ".join(tailcmd), err,
                         error_code=ec.OPERATION_FAILED)
        return result


def get_all_modifiers():
    """retun the list of all the command modifiers"""
    modifier_list = []
    for item in dir(constants):
        if (item.startswith('OUTPUT_MODIFIER') and
                    item != 'OUTPUT_MODIFIER_TOKEN'):
            modifier_list.append(getattr(constants, item))
    return modifier_list
