# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import select
import shlex
import signal
import sys

from cli.descriptors.error import DescriptorException
from cli.utils.central_cli_utils import get_central_cli_nodes
from cli.utils.central_cli_utils import get_cmd_list_path_tuple
from cli.utils.central_cli_utils import is_central_cli_command
from cli.utils.constants import CENTRAL_CLI_END_TOKEN
from cli.utils.constants import CENTRAL_CLI_START_TOKEN
from cli.utils.constants import CENTRAL_CLI_STR
from cli.utils.constants import CLI_DEFAULT_INACTIVITY_TIMEOUT
from cli.utils.constants import END_TOKEN
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_KVM
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_BAREMETAL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.constants import OUTPUT_MODIFIER_COUNT
from cli.utils.constants import OUTPUT_MODIFIER_FIND
from cli.utils.constants import OUTPUT_MODIFIER_FIRST
from cli.utils.constants import OUTPUT_MODIFIER_IGNORE
from cli.utils.constants import OUTPUT_MODIFIER_JSON
from cli.utils.constants import OUTPUT_MODIFIER_LAST
from cli.utils.constants import OUTPUT_MODIFIER_MORE
from cli.utils.constants import OUTPUT_MODIFIER_SORT
from cli.utils.constants import OUTPUT_MODIFIER_TOKEN
from cli.utils.error_utils import render_error
from cli.utils.error_utils import render_error_result
from cli.utils.location_utils import FilePaths
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_windows
from cli.utils.node_utils import should_enable_central_cli
from cli.utils.render_utils import render_definition
from cli.utils.error_code import ErrorCode as ec


_NODE_TYPE = get_node_type()

# The libedit packaged with ESX has proven to be buggy and falling short of
# essential features, which are required for NSXCLI to function. nsx_readline
# packages a pure Python alternative to readline, without any underlying
# system dependencies. This ofcourse comes at a cost of reduced functionality.
# Our hope is that adding any missing feature can be done on need basis, and
# would be easier than fixing and packing modified libedit code.
if _NODE_TYPE == NODE_TYPE_HOST_ESX:
    from cli.utils.nsx_readline import NSXReadline
    readline = NSXReadline()
    compat23_raw_input = readline.raw_input
else:
    import readline  # Don't import readline on ESXi, see #1765653
    # Note, it's possible a supported hypervisor has a libedit based
    # readline installed, if so, use nsx_readline instead
    if readline.__doc__ and 'libedit' in readline.__doc__:
        from cli.utils.nsx_readline import NSXReadline
        readline = NSXReadline()
        compat23_raw_input = readline.raw_input
    else:
        from compat23.compat23_utils import compat23_raw_input


lg = nsx_logging.getLogger("cli.utils.input_utils")


CHAR_COMPLETE = "tab"
if is_windows():
    CHAR_HELP = "shift-?"
else:
    CHAR_HELP = "?"


_NODE_TYPE_HOSTS = [
    NODE_TYPE_HOST_ESX, NODE_TYPE_HOST_KVM, NODE_TYPE_HOST_RHELKVM,
    NODE_TYPE_HOST_UBUNTUKVM, NODE_TYPE_HOST_WIN, NODE_TYPE_HOST_CENTOSKVM,
    NODE_TYPE_HOST_SLESKVM,
    NODE_TYPE_BAREMETAL_SERVER, NODE_TYPE_BAREMETAL_CONTAINER,
    NODE_TYPE_BAREMETAL_UBUNTU_SERVER, NODE_TYPE_BAREMETAL_RHEL_SERVER,
    NODE_TYPE_BAREMETAL_CENTOS_SERVER, NODE_TYPE_BAREMETAL_SLES_SERVER,
    NODE_TYPE_BAREMETAL_WINDOWS_SERVER, NODE_TYPE_BAREMETAL_OEL_SERVER,
    NODE_TYPE_BAREMETAL_RHEL_CONTAINER, NODE_TYPE_BAREMETAL_CENTOS_CONTAINER]

# There is a class of commands where we want to suppress the ambiguous command
# error for usability reasons. These commands take the form of:
#   get objects
#   get object <object-name>  (where <object-name> is a query/enum parameter)
#
# For these commands we want the NSX CLI to show the valid options for
# <object-name> if the user enters: "get object <TAB>".
_SUPPRESS_AMBIGUOUS_CHECK_PATTERNS = [
    "file",       # "get files", "get file <filename>"
    "interface",  # "get interfaces", "get interface <interface-name>"
    "service"     # "get services", "get service <service-name>"
]


# The following values are used to set InputHelper._error_state which is
# only used in unit tests only. When performing tab completion, it is
# necessary to output error messages to the user when it doesn't make
# sense to tab complete. Instead of having unit tests capture stdout to
# check for these errors, InputHelper._error_state is used to check that
# an error was detected
ERROR_STATE_AMBIGUOUS_COMMAND = "AMBIGUOUS_COMMAND"
ERROR_STATE_AMBIGUOUS_MODIFIER = "AMBIGUOUS_MODIFIER"
ERROR_STATE_ARGUMENT_ERROR = "BAD_ERROR"
ERROR_STATE_BAD_ARGUMENT = "BAD_ARGUMENT"
ERROR_STATE_BAD_COMMAND = "BAD_COMMAND"
ERROR_STATE_BAD_MODIFIER = "BAD_MODIFIER"
ERROR_STATE_CENTRAL_CLI_NOT_READY = "CENTRAL_CLI_NOT_READY"
ERROR_STATE_CMD_MODIFIER_BAD_ARG = "CMD_MODIFIER_BAD_ARG"
ERROR_STATE_CMD_MODIFIER_NO_ARGS = "CMD_MODIFIER_NO_ARGS"
ERROR_STATE_CMD_MODIFIER_UNSUPPORTED = "CMD_MODIFIER_UNSUPPORTED"
ERROR_STATE_NO_QUERY_ARG_VALUES = "NO_QUERY_ARG_VALUES"
ERROR_STATE_NON_CENTRAL_CLI_COMMAND = "NON_CENTRAL_CLI_COMMAND"
ERROR_STATE_NON_CENTRAL_CLI_CONTEXT_COMMAND = "NON_CENTRAL_CLI_CONTEXT_COMMAND"
ERROR_STATE_TOO_MANY_MODIFIERS = "TOO_MANY_MODIFIERS"
ERROR_STATE_UNAUTHORIZED_COMMAND = "UNAUTHORIZED_COMMAND"


CANDIDATE_CR = " "


MODIFIER_NUMLINES_ARG = "<number-of-lines>"
MODIFIER_REGEX_ARG = "<regex>"
MODIFIERS = {
    OUTPUT_MODIFIER_COUNT:  MODIFIER_REGEX_ARG,
    OUTPUT_MODIFIER_FIND:   MODIFIER_REGEX_ARG,
    OUTPUT_MODIFIER_FIRST:  MODIFIER_NUMLINES_ARG,
    OUTPUT_MODIFIER_IGNORE: MODIFIER_REGEX_ARG,
    OUTPUT_MODIFIER_JSON:   None,
    OUTPUT_MODIFIER_LAST:   MODIFIER_NUMLINES_ARG,
    OUTPUT_MODIFIER_MORE:   None,
    OUTPUT_MODIFIER_SORT:   None,
}
ALL_MODIFIERS = list(MODIFIERS.keys())
ALL_MODIFIERS.sort()


class InputHelper(object):
    """This class works in conjunction with GNU readline to implement:
        - tab completion
        - double tab completion
        - ? processing
    """
    def __init__(self, prompt, cli_command_service,
                 inactivity_timeout_handler):
        self._prompt = prompt
        self._service = cli_command_service
        self._inactivity_timeout_handler = inactivity_timeout_handler
        self._sig_alarm_set = False
        self._context = ""
        self._candidates = []
        self._display_completions = []
        self._error_state = None
        self._end_key, self._end_help = self._service.get_token_help(END_TOKEN)
        self._enable_central_cli = should_enable_central_cli()
        readline.parse_and_bind("%s: complete" % CHAR_COMPLETE)
        readline.set_completer_delims(' ')
        readline.set_completer(self._completer)
        readline.parse_and_bind("%s: possible-completions" % CHAR_HELP)
        if is_windows():
            readline.set_completion_display_matches_hook(
                self._display_completions_hook_win)
        else:
            readline.set_completion_display_matches_hook(
                self._display_completions_hook)
            if _NODE_TYPE not in _NODE_TYPE_HOSTS:
                signal.signal(signal.SIGALRM, self._inactivity_handler)

    def get_input(self, context):
        """Prompt user for input.

        Args:
            context: Current navigation mode, e.g. enable
        Returns:
            User input.
        """
        self._context = context
        # Set inactivity timeout which expires if the user takes longer than
        # the configured time to enter a command (e.g. "help<enter>"). Note,
        # ideally timeout expires if the user takes too long to enter a key;
        # however, based on limited research there doesn't appear to be a way
        # to hook into raw_input/readline processing to reset the timeout on
        # key press.
        self._set_inactivity_timeout()
        # user_input = compat23_raw_input('xyz >')
        user_input = compat23_raw_input(self.get_prompt())
        self._clear_inactivity_timeout()
        return user_input

    def get_prompt(self):
        return self._prompt.get_prompt(
            session_mode_active=self._service.session_mode_active)

    def _completer(self, user_last_word, completion_index):
        """Callback invoked by readline when user enters completion char."""
        # readline calls this method when completion character(s) are pressed
        try:
            self._clear_inactivity_timeout()
            user_input = readline.get_line_buffer()
            help_requested = self._user_entered_help_char()
            return self._do_completer(
                user_input, completion_index, help_requested)
        except Exception:
            msg = "Error processing readline completion"
            lg.exception(msg, error_code=ec.INTERNAL_ERROR)
            assert False, msg
        finally:
            self._set_inactivity_timeout()
            self._service._clear_remote_arg_cache()

    def _do_completer(self, user_input, completion_index, help_requested=False):
        """Implementation of _completer callback."""

        def _add(target, item, help_text=None):
            target_item = (item, help_text)
            if target_item not in target:
                target.append(target_item)

        def _return_no_candidates():
            self._candidates = []
            return None

        # If completion_index > 0, then self._candidates should already
        # be populated by a previous invocation (when completion_index = 0)
        service = self._service
        if completion_index > 0:
            try:
                return self._candidates[completion_index][0]
            except IndexError:
                # readline keeps incrementing completion_index til None returned
                pass
            # Don't clear self._candidates because readline will still call
            # possible_completions if completion_index is not 0
            return None
        elif not service:
            return _return_no_candidates()
        self._candidates = []
        self._display_completions = []
        self._error_state = None
        # Reject command if more than one output modifier token specified
        num_modifiers = user_input.count(OUTPUT_MODIFIER_TOKEN)
        if num_modifiers > 1:
            pos = user_input.find(OUTPUT_MODIFIER_TOKEN)
            self._too_many_modifiers(user_input[(pos + 1):].strip())
            return None
        input_has_trailing_space = user_input.endswith(" ")
        user_input = user_input.strip()
        user_tokens = shlex.split(user_input.replace(
            OUTPUT_MODIFIER_TOKEN, " %s " % OUTPUT_MODIFIER_TOKEN))
        input_has_modifier_token = False
        modifiers = []
        modifier_has_trailing_space = False
        if num_modifiers == 1:
            # If last word in user input contains the modifier, normalize it;
            # otherwise, the user see weird text replacement due to how
            # readline replaces the last word based on what's returned by
            # this method.
            # - "help|"      -> "help| "
            # - "help|last"  -> "help| last"
            # - "help |last" -> "help | last"
            last_token = shlex.split(user_input)[-1]
            if (not input_has_trailing_space and
                    OUTPUT_MODIFIER_TOKEN in last_token and
                    OUTPUT_MODIFIER_TOKEN != last_token):
                if "%s " % OUTPUT_MODIFIER_TOKEN not in last_token:
                    return last_token.replace(
                        OUTPUT_MODIFIER_TOKEN, "%s " % OUTPUT_MODIFIER_TOKEN)
            input_has_modifier_token = True
            pos = user_tokens.index(OUTPUT_MODIFIER_TOKEN)
            modifiers = user_tokens[(pos + 1):]
            modifier_has_trailing_space = input_has_trailing_space
            input_has_trailing_space = True
            user_tokens = user_tokens[:pos]
        # If current command is a Central CLI command but the factory's
        # node cache isn't ready yet, return retry error
        if (self._enable_central_cli and
            is_central_cli_command(user_tokens) and
                not service._factory.node_cache_available):
            self._central_cli_not_ready()
            return None
        # Set the service's req_dict so that during tab completion processing
        # query and verify args can get context
        req_dict = {"cmd_list": user_tokens, "context": self._context}
        service.set_req_dict(req_dict)
        # Resolve all descriptors that match what user has input before <TAB>
        index = last_token = None
        if input_has_trailing_space:
            index = len(user_tokens)
            last_token = ""
        elif user_tokens:
            index = len(user_tokens) - 1
            last_token = user_tokens[-1]
        try:
            if user_tokens[0] == CENTRAL_CLI_START_TOKEN:
                verify_argument = False
            else:
                verify_argument = True
        except IndexError:  # No input provided
            verify_argument = True

        exact_matches, partial_matches, incomp_matches, value_mismatches = (
            service.resolve_command_descriptor(
                user_tokens, self._context, allow_unauthorized=True,
                verify_argument=verify_argument,
                input_has_trailing_space=input_has_trailing_space))
        exact_desc = None
        matched_desc = None
        modifier_desc = None
        unauthorized_descs = []
        non_central_cli_desc = []
        non_central_cli_context_desc = []
        # Counter for number of enum argument values to check if there are
        # non enum arg values from the matches
        total_enum_arg_values = 0
        # Placeholder variable to check the uniqueness of enum args
        # If None, no enum arg found; If [], more than 1 enum args found;
        # Otherwise there is 1 unique enum arg.
        unique_enum_arg = None
        sort_keys = [] # Key functions, to be used for sorting arg completion

        for desc in exact_matches + partial_matches + incomp_matches:
            if desc.hidden:
                continue
            if not desc.authorized:
                # User is not authorized to invoke current command, keep track
                # of command in case we need to output unauthorized error
                unauthorized_descs.append(desc)
                continue
            if (desc.requires_tty_if_remote and
                    (is_central_cli_command(shlex.split(user_input)) or
                     self._service.session_mode_active) and
                    desc.id != CENTRAL_CLI_STR):
                # User is not allowed to invoke current command using Central
                # CLI, if the command requires tty
                non_central_cli_desc.append(desc)
                continue
            if (desc.context_command and
                    is_central_cli_command(shlex.split(user_input)) and
                    desc.id.endswith(CENTRAL_CLI_STR)):
                # User is not allowed to invoke current command using Central
                # CLI, if the command is context command
                non_central_cli_context_desc.append(desc)
                continue
            # If user tab completes "" (i.e. <TAB> with no input), the list of
            # top level keywords is being requested
            if index is None:
                _add(self._candidates, "%s " % desc.path_tuple[0])
                continue
            # Check for ambiguous commands, i.e. reject user input
            # "get log <TAB>" if the following commands exist:
            # - "get log-file <filename>"
            # - "get logging-servers"
            if matched_desc is None:
                matched_desc = desc
            elif self._check_ambiguous_commands(desc, matched_desc, user_input,
                                                input_has_trailing_space):
                self._ambiguous_command(user_input)
                return _return_no_candidates()
            # If the current command descriptor exactly/partially matches
            # user input, add <CR> and optionally modifier "|" to candidates
            try:
                is_complete = desc.is_command_complete(user_tokens)
                if (is_complete and (index == len(user_tokens)) and
                        (desc in exact_matches or desc in partial_matches)):
                    if desc.additional_last_args:
                        continue
                    _add(self._candidates, CANDIDATE_CR)
                    exact_desc = desc
                    modifier_desc = None
                    if not desc.suppress_modifiers:
                        modifier_desc = desc
                        _add(self._candidates, "%s " % OUTPUT_MODIFIER_TOKEN)
                    continue
            except DescriptorException as exc:
                lg.exception("Failed to process command descriptor",
                             error_code=ec.OPERATION_FAILED)
                continue

            # If we get here, current command descriptor matches what
            # the user has input so far but the command has additional
            # keywords or arguments that need completion. So we add the
            # next keyword/argument in the command as a candidate.
            try:
                item, partial_response = desc.get_command_literal_argument(
                    user_tokens, index)
            except DescriptorException:
                lg.exception("Failed to process command descriptor",
                             error_code=ec.OPERATION_FAILED)
                continue

            is_arg, enum_arg, query_arg, arg_values = (
                service.search_argument_values(item))
            if is_arg:
                if enum_arg or query_arg:
                    if enum_arg:
                        if (unique_enum_arg is None and
                                total_enum_arg_values == 0):
                            # Keep record of the first enum arg encountered
                            unique_enum_arg = arg_values
                            total_enum_arg_values += len(arg_values)
                        elif unique_enum_arg and unique_enum_arg != arg_values:
                            # Set unique_enum_arg to be an empty list if there
                            # are more than one enum args
                            unique_enum_arg = []
                    for val, help_text in arg_values:
                        if val.startswith(last_token):
                            _add(self._candidates, "%s " % val, help_text)
                elif service.is_argument(item) and len(last_token) < 1:
                    _add(self._candidates, "%s " % item)
                elif desc in exact_matches + partial_matches:
                    _add(self._candidates, "%s " % user_tokens[-1])
                arg_desc = service._get_argument_descriptor(item)
                sort_keys.append(arg_desc.sort_key)
            # Add next keyword in command as a candidate
            else:
                if item.startswith(last_token):
                    _add(self._candidates, "%s " % item)

            # Remove items from candidates if already part of user_input
            # TODO: Remove already provided arguments from self._candidates
            # self._candidates = [
            #     x for x in self._candidates
            #     if x[0].strip() not in partial_response]

        if not self._candidates:
            # If no matching commands found but some were rejected because the
            # user is not authorized to invoke them, output unauthorized error
            if unauthorized_descs:
                self._unauthorized_command(user_input)
                return _return_no_candidates()
            # Similarly, if some commands were rejected because they were not
            # allowed to run using Central CLI, output corresponding error
            elif non_central_cli_desc:
                self._non_central_cli_command(user_input)
                return _return_no_candidates()
            # If context command rejected because they are not allowed to run using
            # Central CLI, output corresponding error
            # Note: Context command can be invoke using session mode
            elif non_central_cli_context_desc:
                self._non_central_cli_context_command(user_input)
                return _return_no_candidates()

            # If the user input contained invalid arg values that would have
            # otherwise matched a command, output arg value error
            error_arg = None
            error_arg_value = None
            if value_mismatches:
                for desc in value_mismatches:
                    pos = 0
                    for item in desc.path_tuple:
                        if pos >= len(user_tokens):
                            break
                        val = user_tokens[pos]
                        pos += 1
                        if not service.is_argument(item):
                            continue
                        # Test if a verify or query argument returned an error
                        arg_desc = self._service._get_argument_descriptor(
                            item)
                        if arg_desc.verify_argument:
                            result = arg_desc.match_verify_arg_value(
                                user_input, req_dict, false_if_error=False,
                                debug_mode=service._allow_debug_cmds,
                                remote_nodes=service.session_nodes\
                                    if service.session_mode_active else None)
                        elif arg_desc.query_argument:
                            result = arg_desc.get_allowed_values(
                                req_dict,
                                empty_list_if_error=False,
                                debug_mode=service._allow_debug_cmds,
                                session_nodes=service.session_nodes)
                        else:
                            continue
                        if isinstance(result, str):
                            self._argument_error(result)
                            return _return_no_candidates()
                        if service.match_argument_value(item, val) is None:
                            error_arg = item
                            error_arg_value = val
                            break
                    if error_arg:
                        break
            elif len(incomp_matches) == 1:
                # Check if tab completion failed because
                # a query argument returned no valid values
                desc = incomp_matches[0]
                tmp_user_tokens = get_cmd_list_path_tuple(user_tokens)
                if len(desc.path_tuple) > len(tmp_user_tokens):
                    item = desc.path_tuple[len(tmp_user_tokens)]
                    if service.is_argument(item):
                        arg_desc = service._get_argument_descriptor(item)
                        if arg_desc is not None and arg_desc.query_argument:
                            self._no_query_argument_values(item)
                            return _return_no_candidates()
            self._bad_command(error_arg, error_arg_value, user_tokens)
            return _return_no_candidates()
        # If user input specified a modifier token ("|"), check if a matching
        # command descriptors was found that accepts modifiers.
        if input_has_modifier_token:
            if not exact_desc:
                self._bad_command(None, None, user_tokens)
                return _return_no_candidates()
            elif not modifier_desc:
                self._command_modifier_unsupported()
                return _return_no_candidates()
            self._do_modifier_completer(modifiers, modifier_has_trailing_space)
        # Sort all candidates if there is any non enum arg or more than one
        # unique enum args
        if (total_enum_arg_values != len(self._candidates) or
                not unique_enum_arg):
                # Use the sort_key method from argument descriptor only if there
                # is only one such value. Otherwise, perform default sorting.
                self._candidates.sort(key=sort_keys[0]
                                      if len(sort_keys) == 1 else None)
        # Perform special case handling when there's only 1 candidate:
        # 1. If the candidate is (CANDIDATE_CR, *), show help for the candidate
        #    and return no candidates; otherwise, readline keeps appending an
        #    empty space on each <TAB>.
        # 2. If help_requested is True (i.e. user entered <?>), show help
        #    for the candiate and return no candidates; otherwise, readline
        #    outputs the candidate on a new line to stdout and we have no way
        #    of formatting the output.
        # 3. If the candidate is an argument name, e.g. <arg_name>, show help
        #    for the candidate and return no candidates; otherwise, readline
        #    outputs the candidate on the current line to stdout.
        if len(self._candidates) == 1:
            candidate = self._candidates[0][0]
            if ((candidate == CANDIDATE_CR) or
                    help_requested or
                    service.is_argument(candidate.strip())):
                self._display_completions_found()
                return _return_no_candidates()
        # If all the candidates are arguments, show completions using
        # display_completions_found; otherwise, readline will try to complete
        # with "<..." in the user input line.
        elif all(
                [service.is_argument(c[0].strip()) for c in self._candidates]):
            self._display_completions_found()
            return _return_no_candidates()

        return self._candidates[0][0] if self._candidates else None

    def _do_modifier_completer(self, modifiers, modifier_has_trailing_space):
        self._candidates = []
        if not modifiers:
            if not modifier_has_trailing_space:
                self._candidates = [("%s " % OUTPUT_MODIFIER_TOKEN, None)]
            else:
                for modifier in ALL_MODIFIERS:
                    self._candidates.append(("%s " % modifier, None))
            return self._candidates[0][0]
        matched_modifiers = []
        for modifier in ALL_MODIFIERS:
            if modifier.startswith(modifiers[0]):
                matched_modifiers.append(modifier)
        # If not matching modifiers found or if user input contained more
        # than 2 modifier keywords (currently no modifier accepts more than
        # 2 keywords), return error
        if not matched_modifiers or len(modifiers) > 2:
            self._bad_modifier(modifiers)
            return None
        # If only one modifier keyword specified, complete it
        if len(modifiers) == 1:
            if not modifier_has_trailing_space:
                self._candidates = [("%s " % m, None) for m in matched_modifiers]
                return self._candidates[0][0]
            # User has specified "MODIFER <TAB>", complete the modifier
            # argument if the modifier accepts an argument. But first make
            # sure modifier is not ambiguous (e.g. "find" vs "first")
            if len(matched_modifiers) > 1:
                self._ambiguous_modifier(modifiers[0], matched_modifiers)
                return None
            for modifier in matched_modifiers:
                modifier_arg = MODIFIERS.get(modifier)
                if modifier_arg:
                    self._candidates = [("%s " % modifier_arg, None)]
                    return self._candidates[0][0]
                else:
                    return None
        # Two modifier keywords specified, make sure:
        # 1. modifier is not ambiguous
        # 2. modifier accepts arguments
        # 3. modifier arg is of the right type
        if len(matched_modifiers) > 1:
            self._ambiguous_modifier(modifiers[0], matched_modifiers)
            return None
        modifier_arg = MODIFIERS.get(matched_modifiers[0])
        if not modifier_arg:
            self._modifier_does_not_accept_args(modifiers[0])
            return None
        elif modifier_arg == MODIFIER_NUMLINES_ARG:
            try:
                _ = int(modifiers[1])
            except ValueError:
                self._invalid_modifier_arg(MODIFIER_NUMLINES_ARG, modifiers[1])
                return None
        if modifier_has_trailing_space:
            # Tab preceded by a space, nothing to complete
            return None
        self._candidates = [("%s " % modifiers[1], None)]
        return self._candidates[0][0]

    def _display_completions_hook(self, unused_substitution, unused_matches,
                                  unused_longest):
        """Callback invoked by readline to show completions for a command."""
        # readline calls this method when there are more than one possible
        # completions or if CHAR_HELP is entered
        self._display_completions_found()

    def _display_completions_hook_win(self, unused_matches):
        """Callback invoked by readline to show completions for a command."""
        # readline calls this method when there are more than one possible
        # completions or if CHAR_HELP is entered
        self._display_completions_found()

    def _display_completions_found(self):
        """Print possible completions stored in self._candidates."""
        # Set self._display_completions so we can unit test what
        # completion options are displayed to the user
        self._display_completions = self._candidates
        end_token_exists = (CANDIDATE_CR, None) in self._candidates
        output_modifier_token_exists = (
            ("%s " % OUTPUT_MODIFIER_TOKEN, None) in self._candidates)
        max_len = 8
        options = {}
        sorted_keys = []
        for item, help_text in self._candidates:
            key, tmp_help_text = self._service.get_token_help(item.strip())
            help_text = tmp_help_text if help_text is None else help_text
            options[key] = help_text or key
            # Do not add duplicate key in sorted_keys list
            if key not in sorted_keys:
                sorted_keys.append(key)
            max_len = len(key) if len(key) > max_len else max_len

        # If Central CLI command, remove already specified completions. This is
        # necessary as node argument is a multi-arg. After that, make sure that
        # CENTRAL_CLI_END_TOKEN keyword is always the last in completions.
        line_buffer_tuple = shlex.split(readline.get_line_buffer())
        if is_central_cli_command(line_buffer_tuple):
            nodes = get_central_cli_nodes(line_buffer_tuple)
            #  No nodes were specified, remove CENTRAL_CLI_END_TOKEN
            if not nodes:
                sorted_keys.remove(CENTRAL_CLI_END_TOKEN)
            else:
                for n in nodes:
                    try:
                        sorted_keys.remove(n)
                    except ValueError:
                        pass  # Ignore removing keys which don't exist

                try:  # Ensure CENTRAL_CLI_END_TOKEN is last candidate
                    sorted_keys.remove(CENTRAL_CLI_END_TOKEN)
                    sorted_keys.append(CENTRAL_CLI_END_TOKEN)
                except ValueError:  # CENTRAL_CLI_END_TOKEN not in candidates
                    pass

        # Always output end token (<cr>) second last
        if end_token_exists:
            sorted_keys.remove(self._end_key)
            sorted_keys.append(self._end_key)
        # Always output output modifier token (|) last
        if output_modifier_token_exists:
            sorted_keys.remove(OUTPUT_MODIFIER_TOKEN)
            sorted_keys.append(OUTPUT_MODIFIER_TOKEN)
        # Incase there are no candidates, shift to new line first
        if not self._candidates:
            sys.stdout.write("\n")
        if sorted_keys:
            sys.stdout.write("\n")
            for key in sorted_keys:
                fmt = "  %%-%ds  %%-s\n" % max_len
                sys.stdout.write(fmt % (key, options[key]))
            sys.stdout.write("\n")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _check_ambiguous_commands(self, new_desc, matched_desc,
                                  user_input, input_has_trailing_space):
        """Return True if the specified cli_command_descriptors
           result in an ambiguous command condition."""
        command_indexes = new_desc.get_command_indexes(shlex.split(user_input))
        if input_has_trailing_space:
            index = len(command_indexes)
        else:
            index = len(command_indexes) - 1

        same_start = True
        len_new = len(new_desc.path_tuple)
        len_matched = len(matched_desc.path_tuple)

        resolve_argument_name = lambda _str:\
                                    _str[1:-1] if _str.startswith("<") and\
                                        _str.endswith(">")\
                                    else _str

        for pos in range(index):

            # If new_desc's current element ['current_arg'] @ `pos` index is of argument type.
            if self._service.is_argument(matched_desc.path_tuple[pos]):

                # From the argument name of current element ['current_arg'],
                # find it's descriptor to verify whether new_desc's values are subset of matched_desc.
                # PR: 2499431, is a scenario where cmds starting with -
                #     > get logical-router <nsxa-uuid-service-router-arg> high-availability ....
                #     > get logical-router <dpd-uuid-lrouter-arg> ....
                # Whereas the values of argument `nsxa-uuid-service-router-arg` is complete subset of
                # argument `dpd-uuid-lrouter-arg`.
                arg_desc = self._service._get_argument_descriptor(new_desc.path_tuple[pos])

                # Such related argument descriptors are mapped using `is_subset_of` yaml property.
                # incase of related argument values, cmds will not be ambiguous till this point of comparision.
                if resolve_argument_name(matched_desc.path_tuple[pos])\
                   in arg_desc.is_subset_of:
                    continue

            if (pos >= len_new or pos >= len_matched or
                    new_desc.path_tuple[pos] != matched_desc.path_tuple[pos]):
                same_start = False
                break
        if same_start:
            return False
        # If we got here, new_desc and matched_desc result in an
        # ambiguous command condition. However, for usability reasons,
        # we check to see if the two descriptors are have the pattern:
        # "get objects" and "get object <object-name>"
        # We want to allow "get object <TAB>" to list object-names
        get_objects_desc = None
        get_object_arg_desc = None

        for desc in [new_desc, matched_desc]:
            path_tuple = desc.path_tuple
            if (not (path_tuple[0] == "get" or
                    (is_central_cli_command(path_tuple) and
                     path_tuple[2] == "get"))):
                continue
            elif (len(path_tuple) == 2 and
                    path_tuple[1].endswith("s") and
                    path_tuple[1][:-1] in _SUPPRESS_AMBIGUOUS_CHECK_PATTERNS):
                get_objects_desc = desc
            elif (len(path_tuple) == 4 and
                    path_tuple[0] == CENTRAL_CLI_START_TOKEN and
                    path_tuple[3].endswith("s") and
                    path_tuple[3][:-1] in _SUPPRESS_AMBIGUOUS_CHECK_PATTERNS):
                get_objects_desc = desc
            elif (len(path_tuple) == 3 and
                    path_tuple[1] in _SUPPRESS_AMBIGUOUS_CHECK_PATTERNS and
                    self._service.is_argument(path_tuple[2])):
                get_object_arg_desc = desc
            elif (len(path_tuple) == 5 and
                    path_tuple[3] in _SUPPRESS_AMBIGUOUS_CHECK_PATTERNS and
                    self._service.is_argument(path_tuple[4])):
                get_object_arg_desc = desc
        if get_objects_desc is None or get_object_arg_desc is None:
            return True
        return False

    def _bad_command(self, error_arg, error_arg_value, user_tokens):
        """Print invalid argument error or generic unknown command error."""
        sys.stdout.write("\n")
        if error_arg and error_arg_value:
            self._error_state = ERROR_STATE_BAD_ARGUMENT
            arg_label, _ = self._service.get_token_help(error_arg)
            render_error("MSG_CLI_INVALID_ARGUMENT", arg_label, error_arg_value)
        else:
            self._error_state = ERROR_STATE_BAD_COMMAND
            self._service.command_not_found(
                " ".join(user_tokens), context=self._context)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _unauthorized_command(self, user_input):
        """Print unauthorized command error."""
        self._error_state = ERROR_STATE_UNAUTHORIZED_COMMAND
        sys.stdout.write("\n")
        # The specified input is unauthorized, show alternatives
        parts = shlex.split(user_input)
        hint = "%s %s" % (parts[0], parts[1][0]) if len(parts) > 1 else parts[0]
        self._service.ambiguous_command_found(
            user_input, context=self._context, suppress_hidden_debug=True,
            search_hint=hint, custom_error_id="MSG_CLI_COMMAND_NOT_AUTHORIZED",
            custom_alt_error_id="MSG_CLI_COMMAND_NOT_AUTHORIZED_VALUE_ALT")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _ambiguous_command(self, user_input):
        """Print ambiguous command error."""
        self._error_state = ERROR_STATE_AMBIGUOUS_COMMAND
        sys.stdout.write("\n")
        self._service.ambiguous_command_found(
            user_input, context=self._context,
            suppress_hidden_debug=True)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _ambiguous_modifier(self, input_modifier, possible_modifiers):
        """Print ambiguous modifier error."""
        self._error_state = ERROR_STATE_AMBIGUOUS_MODIFIER
        sys.stdout.write("\n")
        render_error(
            "MSG_CLI_OUTPUT_MODIFIER_MULTIPLE_MATCHED", input_modifier,
            " or ".join(possible_modifiers))
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _bad_modifier(self, modifiers):
        """Print invalid modifier error."""
        self._error_state = ERROR_STATE_BAD_MODIFIER
        sys.stdout.write("\n")
        msg_id = "MSG_CLI_OUTPUT_MODIFIER_INVALID"
        render_error(msg_id,  " ".join(modifiers))
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _too_many_modifiers(self, modifier):
        """Print too many modifiers error."""
        self._error_state = ERROR_STATE_TOO_MANY_MODIFIERS
        sys.stdout.write("\n")
        render_error("MSG_CLI_OUTPUT_MODIFIER_EXCEEDED", modifier)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _command_modifier_unsupported(self):
        self._error_state = ERROR_STATE_CMD_MODIFIER_UNSUPPORTED
        sys.stdout.write("\n")
        render_error("MSG_CLI_CMD_DOES_NOT_SUPPORT_OUTPUT_MODIFIER")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _modifier_does_not_accept_args(self, input_modifier):
        """Print modifier does not accept arguments error."""
        self._error_state = ERROR_STATE_CMD_MODIFIER_NO_ARGS
        sys.stdout.write("\n")
        render_error(
            "MSG_CLI_OUTPUT_MODIFIER_NO_ARGS", input_modifier)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _invalid_modifier_arg(self, arg_name, arg_value):
        """Print invalid modifier argument error."""
        self._error_state = ERROR_STATE_CMD_MODIFIER_BAD_ARG
        sys.stdout.write("\n")
        render_error(
            "MSG_CLI_OUTPUT_MODIFIER_INVALID_ARG", arg_value)
        arg_desc = self._service._get_argument_descriptor(arg_name)
        if arg_desc and arg_desc.description:
            render_definition(arg_desc.label, arg_desc.description, False)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _argument_error(self, error_details):
        """Print argument error returned by argument descriptor."""
        self._error_state = ERROR_STATE_ARGUMENT_ERROR
        sys.stdout.write("\n")
        render_error_result(error_details)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _no_query_argument_values(self, error_arg):
        sys.stdout.write("\n")
        self._error_state = ERROR_STATE_NO_QUERY_ARG_VALUES
        arg_label, _ = self._service.get_token_help(error_arg)
        render_error("MSG_CLI_NO_VALUES_FOR_QUERY_ARGUMENT", arg_label)
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _non_central_cli_command(self, user_input):
        """Print non Central CLI command error."""
        self._error_state = ERROR_STATE_NON_CENTRAL_CLI_COMMAND
        sys.stdout.write("\n")
        render_error("MSG_CLI_CMD_DOES_NOT_SUPPORT_CENTRAL_CLI")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _non_central_cli_context_command(self, user_input):
        """Print not central CLI context command error."""
        self._error_state = ERROR_STATE_NON_CENTRAL_CLI_CONTEXT_COMMAND
        sys.stdout.write("\n")
        render_error("MSG_CLI_CMD_REQUIRES_SESSION_MODE")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _central_cli_not_ready(self):
        """Print system not ready for Central CLI commands."""
        self._error_state = ERROR_STATE_CENTRAL_CLI_NOT_READY
        sys.stdout.write("\n")
        render_error("MSG_CLI_CENTRAL_CLI_STILL_LOADING")
        if (not is_windows()):
            sys.stdout.write("%s%s" % (self.get_prompt(),
                             readline.get_line_buffer()))
            sys.stdout.flush()

    def _inactivity_handler(self, signum, frame):
        """Callback for signal.SIGALRM."""
        rlist, _, _ = select.select([sys.stdin], [], [], 1)
        if rlist == []:
            self._inactivity_timeout_handler()
        else:
            self._clear_inactivity_timeout()
            self._set_inactivity_timeout()

    def _set_inactivity_timeout(self):
        """Start inactivity timeout clock."""
        if (self._inactivity_timeout_handler is None or
                _NODE_TYPE in _NODE_TYPE_HOSTS):
            return
        # Reread the inactivity timeout each time in case the configured
        # value has changed while a CLI session is active
        inactivity_timeout = self._service.cli_config.get('cli_timeout')
        if inactivity_timeout is None:
            inactivity_timeout = CLI_DEFAULT_INACTIVITY_TIMEOUT
        if inactivity_timeout > 0:
            try:
                signal.alarm(inactivity_timeout)
                self._sig_alarm_set = True
            except Exception:
                lg.exception("Error setting inactivity timeout",
                             error_code=ec.INTERNAL_ERROR)

    def _clear_inactivity_timeout(self):
        """Cancel inactivity timeout clock."""
        if self._sig_alarm_set:
            self._sig_alarm_set = False
            signal.alarm(0)

    def _user_entered_help_char(self):
        """Return True if user has requested help via ? or <TAB><TAB>"""
        if (is_windows()):
            return str(readline.get_completion_type()) == CHAR_HELP
        else:
            # On Unix, readline.get_completion_type() returns an int
            return chr(readline.get_completion_type()) == CHAR_HELP
