# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
from functools import partial
import nsx_logging
from mako.template import Template
import os
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import struct
from subprocess import Popen
from subprocess import PIPE
import sys
import datetime

from cli.utils.constants import OUTPUT_MODIFIER_COUNT
from cli.utils.constants import OUTPUT_MODIFIER_FIND
from cli.utils.constants import OUTPUT_MODIFIER_FIRST
from cli.utils.constants import OUTPUT_MODIFIER_IGNORE
from cli.utils.constants import OUTPUT_MODIFIER_MORE
from cli.utils.constants import OUTPUT_MODIFIER_JSON
from cli.utils.constants import OUTPUT_MODIFIER_LAST
from cli.utils.constants import OUTPUT_MODIFIER_SORT
from cli.utils.constants import TEMPFILE_PREFIX
from cli.utils.error_utils import get_message
from cli.utils.node_utils import is_windows
from cli.utils.system_utils import get_system_datetime
from cli.utils.error_code import ErrorCode as ec
from compat23.compat23_utils import compat23_unicode
if not is_windows():
    import fcntl
    import termios


lg = nsx_logging.getLogger("cli.utils.render_utils")


_CMD_GREP = "/bin/grep"
_CMD_GREP_COUNT = "/bin/grep --count"
_CMD_GREP_IGNORE = "/bin/grep -v"
_CMD_LESS = "/usr/bin/less -X -d -F -K"
_CMD_SORT = "/usr/bin/sort --ignore-case"
_CMD_HEAD = "/usr/bin/head"
_CMD_TAIL = "/usr/bin/tail"

# Popen with universal_newlines set to True decodes the data with default
# encoding and always returns 'str' by default from communicate call (It's 'str'
# by default in Python 2.x, but 'byte' in Python 3.x.), making explicit
# decode('utf') obsolete. It's compatible with both Python 2.x and Python 3.x.
# Also, all line termination characters like '\r', '\r\n' and '\n' are seen as
# '\n' by Python program.
Popen = partial(Popen, universal_newlines=True)


is_py3 = lambda: sys.version_info >= (3,)


def get_text_pager():
    """Return pager object that pages (forwards/backwards) user supplied text
       specified via the pager.write() method.

    Return:
        Pager object or None if an error occurred creating the pager.
    Notes:
        1. Caller must call pager.close() when the pager is no longer needed.
        2. The pager can raise IOError and KeyboardInterrupt when the user
           issues cntl-C or quits the pager. These exceptions should be
           ignored.
    """
    os.putenv("LESSSECURE", "1")
    # Using os.popen here instead of subprocess.Popen because the
    # latter disable text highlighting when the user performs a
    # a search.
    try:
        return os.popen(_CMD_LESS, "w")
    except Exception:
        lg.exception("Error invoking: %s", _CMD_LESS,
                     error_code=ec.INTERNAL_ERROR)
        return None


def as_unicode(dct):
    """Hook to encode any unicode string as UTF-8, within a dictionary"""
    if sys.version_info < (3,):
        for key in list(dct.keys()):
            if type(dct[key]) is compat23_unicode:
                dct[key] = dct[key].encode('utf-8')

    return dct


def render_mako_template_str(template_str, provider_output,
                             modifier_str, sort_key=None, output_prefix=None):
    """Print provider output using specifed mako template and output modifier.

    Args:
        template_str: mako template used to render provider output.
        provider_output: Output from provider implementation.
        modifer_str: Output modifier, if any, specified by user.
        sort_key: key param of python list's sort function.
                  >> li.sort(key=sort_key)
        output_prefix: value is prefixed with rendered output data.
    """
    data_sorted = False
    if sys.version_info < (3,):
        # Remove any characters, which may have not been correctly encoded
        provider_output = provider_output.decode("utf-8", errors="ignore")
    # The following line is what the user sees when a provider result
    # is rendered
    if modifier_str and modifier_str.startswith(OUTPUT_MODIFIER_JSON):
        # Output should already be in a json formatted string,
        # but we pretty print it before outputing
        output = simplejson.dumps(simplejson.loads(provider_output),
                    ensure_ascii=False, indent=2, sort_keys=True)
        modifier_str = None
    else:
        # The default_filters of 'decode.utf8' will programatically replace the
        # usual unicode/str function with a specific encoding. This essentially
        # means that templates don't need to explicitly prefix any string with
        # 'u', to treat it as unicode string in Python 2.x
        template = Template(template_str, default_filters=['decode.utf8'])
        # Mako template's render method returns unicode, we convert to UTF-8
        try:
            obj = simplejson.loads(provider_output, object_hook=as_unicode)
            if modifier_str and modifier_str.startswith(OUTPUT_MODIFIER_SORT)\
               and hasattr(obj, 'sort') and sort_key:
                obj.sort(key=sort_key)
                data_sorted = True
        except Exception:
            lg.exception(
                "Error loading JSON output: %s", provider_output,
                error_code=ec.INTERNAL_ERROR)
            raise
        output = template.render(result=obj).rstrip().encode("utf-8")
        if sys.version_info >= (3,):
            output = output.decode("utf8")
    if output_prefix:
        # Inorder to keep output intact in stdout file; as it may be consumed by sub-processes,
        # hence output_prefix should be displayed on stderr.
        sys.stderr.write(output_prefix)
        sys.stderr.write("\n")
        sys.stderr.flush()
    if output:
        # Due to control logic (usually at the beginning of a mako template),
        # the output may start with blank lines, filter these out
        while True:
            if not output.startswith("\n"):
                break
            output = output[1:]
        _render_formatted_output(output, modifier_str, data_sorted)
        sys.stdout.write("\n")  # empty line before showing prompt


def render_definition(token, definition, add_blank_line=True):
    """Print command token and its definition (called when user enters ?).

    Args:
        token: A command token to print.
        definition: The command token's definition to print.
        add_blank_line: If True, add extra blank line after definition line.
    """
    sys.stdout.write("%s: %s\n" % (token, definition))
    if add_blank_line:
        sys.stdout.write("\n")  # empty line before showing prompt


def _render_formatted_output(formatted_output, modifier_str,
                             data_sorted=False):
    """Print mako formatted output accounting for output modifier if specified.

    Args:
        formatted_output: Output after applying mako template to print.
        modifier_str: Output modifier, if any, specified by user.
        data_sorted: flag; when enabled specifies rendered data already sorted.
    """
    fmt = arg = None
    lines = formatted_output.split("\n")
    if modifier_str:
        parts = modifier_str.split()
        fmt = parts[0]
        arg = " ".join(parts[1:])
    if fmt == OUTPUT_MODIFIER_COUNT:
        count = 0
        for line in lines:
            if re.search(arg, line):
                count += 1
        sys.stdout.write(
            "%s\n" % get_message("MSG_CLI_MATCHED_LINES", arg, count))
        return
    elif fmt == OUTPUT_MODIFIER_MORE:
        pager = get_text_pager()
        if pager:
            try:
                pager.write(formatted_output)
            except (IOError, KeyboardInterrupt):
                pass
            finally:
                pager.close()
        return
    elif fmt == OUTPUT_MODIFIER_FIND:
        lines = [line for line in lines if re.search(arg, line)]
    elif fmt == OUTPUT_MODIFIER_IGNORE:
        lines = [line for line in lines if not re.search(arg, line)]
    elif fmt == OUTPUT_MODIFIER_SORT and not data_sorted:
        lines = sorted(lines, key=lambda line: line.lower())
    elif fmt == OUTPUT_MODIFIER_FIRST:
        try:
            num = int(arg)
            lines = lines[:num] if num else []
        except ValueError:
            # CLI framework should have validated arg,
            # so we should never get here
            pass
    elif fmt == OUTPUT_MODIFIER_LAST:
        try:
            num = int(arg)
            lines = lines[-int(arg):] if num else []
        except ValueError:
            # CLI framework should have validated arg,
            # so we should never get here
            pass
    try:
        for line in lines:
            sys.stdout.write("%s\n" % line)
    except KeyboardInterrupt:
        pass
    return


def format_subprocess_output(subprocess_cmd_list, modifier_str=None,
                             less_output=False):
    """Apply modifier_str to subprocess output and print result to stdout.

    Args:
        subprocess_cmd_list: Command list to invoke using subprocess.
        modifier_str: Output modifier, if any, specified by user to apply.
        less_output: Redirect output to /usr/bin/less.
    Notes:
        1) modifier_str, if specified, should only start with "count",
           "find", "ignore", "more" or "sort". If it starts with a
           different word, this method returns immediately.
        2) If modifier_str is specified and starts with "count" or "more"
           then less_output is assumed False, (e.g. we don't want to
           redirect output to less twice or when counting).
    """
    # If modifier specified, construct the corresponding modifier command
    modifier_cmd = None
    if modifier_str:
        modifier_parts = modifier_str.split()
        if modifier_parts[0] not in [
                OUTPUT_MODIFIER_COUNT, OUTPUT_MODIFIER_FIND,
                OUTPUT_MODIFIER_IGNORE, OUTPUT_MODIFIER_MORE,
                OUTPUT_MODIFIER_SORT, OUTPUT_MODIFIER_FIRST,
                OUTPUT_MODIFIER_LAST]:
            lg.error("Invalid modifier specified: %s", modifier_parts[0],
                     error_code=ec.INVALID_ARGUMENT)
            return
        if modifier_parts[0] == OUTPUT_MODIFIER_FIRST:
            modifier_parts[1] = "-%s" % modifier_parts[1]
            modifier_cmd = _CMD_HEAD.split() + modifier_parts[1:]
        elif modifier_parts[0] == OUTPUT_MODIFIER_LAST:
            modifier_parts[1] = "-%s" % modifier_parts[1]
            modifier_cmd = _CMD_TAIL.split() + modifier_parts[1:]
        elif modifier_parts[0] == OUTPUT_MODIFIER_COUNT:
            less_output = False
            modifier_cmd = _CMD_GREP_COUNT.split() + modifier_parts[1:]
        elif modifier_parts[0] == OUTPUT_MODIFIER_FIND:
            modifier_cmd = _CMD_GREP.split() + modifier_parts[1:]
        elif modifier_parts[0] == OUTPUT_MODIFIER_IGNORE:
            modifier_cmd = _CMD_GREP_IGNORE.split() + modifier_parts[1:]
        elif modifier_parts[0] == OUTPUT_MODIFIER_MORE:
            less_output = False
            modifier_cmd = _CMD_LESS.split()
            os.putenv("LESSSECURE", "1")
        elif modifier_parts[0] == OUTPUT_MODIFIER_SORT:
            modifier_cmd = _CMD_SORT.split()
    cmd_po = modifier_po = less_po = None
    try:
        cmd_stdout = PIPE if modifier_cmd or less_output else None
        cmd_po = Popen(subprocess_cmd_list, stdout=cmd_stdout, stderr=PIPE)
        if modifier_cmd:
            modifier_stdout = (
                PIPE if less_output or
                modifier_parts[0] == OUTPUT_MODIFIER_COUNT else None)
            modifier_po = Popen(
                modifier_cmd, stdin=cmd_po.stdout, stdout=modifier_stdout,
                stderr=PIPE)
            cmd_po.stdout.close()
        if less_output:
            less_stdin = modifier_po.stdout if modifier_cmd else cmd_po.stdout
            less_po = Popen(_CMD_LESS.split(), stdin=less_stdin, stderr=PIPE)
            less_stdin.close()
    except KeyboardInterrupt:
        pass
    except Exception:
        msg = "Error invoking: %s" % " ".join(subprocess_cmd_list)
        if modifier_str:
            msg = "%s | %s" % (msg, modifier_str)
        lg.exception(msg, error_code=ec.INTERNAL_ERROR)
    finally:
        if less_po:
            err = None
            try:
                _, err = less_po.communicate()
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            if err:
                lg.error("Error invoking: command | less, err: %s", err,
                         error_code=ec.OPERATION_FAILED)
        if modifier_po:
            if less_output:
                modifier_po.kill()
            out = err = None
            try:
                out, err = modifier_po.communicate()
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            except ValueError:  # I/O operation on closed file
                pass
            if modifier_parts[0] == OUTPUT_MODIFIER_COUNT and out:
                sys.stdout.write("%s\n" % get_message(
                    "MSG_CLI_MATCHED_LINES", " ".join(modifier_parts[1:]), out))
            if err:
                lg.error(
                    "Error invoking: %s | %s, err: %s",
                    " ".join(subprocess_cmd_list), modifier_str, err,
                    error_code=ec.OPERATION_FAILED)
        if cmd_po:
            # TODO: For log files that are readonly by root, we're not able to
            # Popen.kill the process (because the CLI is running as the admin
            # user). So for now, we force the user to enter Ctrl-C. When time
            # permits, implement a sudo/setuid script to kill
            #if modifier_cmd or less_output:
            #    cmd_po.kill()
            #else:
            #    cmd_po.wait()
            err = None
            try:
                _, err = cmd_po.communicate()
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            except ValueError:  # I/O operation on closed file
                pass
            if err:
                lg.error(
                    "Error invoking: %s, err: %s",
                    " ".join(subprocess_cmd_list), err,
                    error_code=ec.OPERATION_FAILED)


def terminal_size():
    """On Non-Windows, return terminal (#rows, #columns) tuple; else (0, 0)"""
    if not is_windows():
        try:
            h, w, hp, wp = struct.unpack(
                'HHHH', fcntl.ioctl(0, termios.TIOCGWINSZ,
                                    struct.pack('HHHH', 0, 0, 0, 0)))
            return h, w
        except IOError:
            lg.exception("Unable to determine terminal size",
                         error_code=ec.IO_ERROR)
    return 0, 0


def generate_datetime_stamp_str(time=None, format=None):
    """
    Generates datetime format string
    Args:
        time: int/float time.time() value,
            if None, generates current datetime stamp value.
        format: strftime format string
    """
    if is_py3() and time is not None:
        if format is None:
            # Fomart as per nsx-cli's "get clock" command
            format = "%a %b %d %Y %Z %H:%M:%S.%f"
        date_obj = datetime.datetime.utcfromtimestamp(time)
        date_obj = date_obj.replace(tzinfo=datetime.timezone.utc).astimezone()
        return date_obj.strftime(format)[0:-3]
    else:
        # Incase of python2 datetime module doesn't have 'timezone' module
        # hence, fallback to default system mechanism to get current system time.
        result = get_system_datetime()
        if result is not None:
            return "%s %s %s" % (
                result.get("date", ""), result.get("timezone", ""),
                result.get("time", ""))
        return ""
