# ***************************************************************************
# Copyright 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
"""This utility script will read central node config json via stdin and then
update configurations for NTP, Syslog, and SNMP accordingly

- The script returns exit code 0 on success and 1 on error
- If exit code is 1, a single line error message will be written to stderr
- If errors are encountered for one of the configurations, the script will
continue to update the other configurations

Example input JSON:

    {
        "timezone": "UTC",
        "ntp": {
            "servers": [
                "ntp1.vmware.com",
                "ntp2.vmware.com"
            ]
        },
        "syslog": {
            "exporters": [
                {
                    "server": "loginsight.vmware.com",
                    "port": 514,
                    "level": "INFO",
                    "protocol": "LI",
                },
                {
                    "server": "1.1.1.1",
                    "port": 515,
                    "level": "INFO",
                    "protocol": "UDP",
                }
            ]
        },
        "snmp": {
            "communities": [
                {
                    "access": "read_only",
                    "community_string": "public3"
                },
                {
                    "access": "read_only",
                    "community_string": "public4"
                }
            ],
            "v2_targets": [
                {
                    "port": 1162,
                    "server": "10.10.10.10",
                    "community_string": "public1"
                },
                {
                    "port": 1162,
                    "server": "10.10.10.11",
                    "community_string": "public2"
                }
            ],
            "v3_auth_protocol": "SHA1",
            "v3_priv_protocol": "AES128",
            "v3_targets": [
                {
                    "port": 1162,
                    "security_level": "AUTH_PRIV",
                    "server": "10.10.10.10",
                    "user_id": "nsx-collector2"
                },
                {
                    "port": 1162,
                    "security_level": "AUTH_PRIV",
                    "server": "10.10.10.11",
                    "user_id": "nsx-collector1"
                }
            ],
            "v3_users": [
                {
                    "access": "READ_ONLY",
                    "security_level": "AUTH_PRIV",
                    "user_id": "nsx-collector1",
                    "auth_password":"VMwarensbu_1",
                    "priv_password":"VMwarensbu_1"
                },
                {
                    "access": "READ_ONLY",
                    "security_level": "AUTH_PRIV",
                    "user_id": "nsx-collector2",
                    "auth_password":"VMwarensbu_2",
                    "priv_password":"VMwarensbu_2"
                }
            ]
        }
    }

"""

import json
import sys
import subprocess
import uuid
import traceback

sys.path.append("/opt/vmware/nsx-common/python")
import nsx_logging
from nsx_utils.node_utils import get_node_type
from nsx_utils.node_utils import is_appliance
from nsx_utils.constants import NODE_TYPE_HOST_CENTOSKVM
from nsx_utils.constants import NODE_TYPE_HOST_SLESKVM
from nsx_utils.constants import NODE_TYPE_HOST_RHELKVM
from nsx_utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from nsx_utils.snmp_utils import parse_snmp_configs
from nsx_utils.snmp_utils import get_snmp_configs
from nsx_utils.snmp_utils import validate_snmp_config
from nsx_utils.snmp_utils import update_snmp_config

_NODE_TYPE = get_node_type()

_PYTHON_VERSION = '3' if sys.version_info >= (3,) else '2'

_NAPI_PREFIX = "http://localhost:7441/api/v1"
_NODE_URI = "%s/node" % _NAPI_PREFIX
_NTP_SERVICE_URI = "%s/node/services/ntp" % _NAPI_PREFIX
_SYSLOG_EXPORTERS_URI = "%s/node/services/syslog/exporters" % _NAPI_PREFIX
_SNMP_SERVICE_URI = "%s/node/services/snmp" % _NAPI_PREFIX

_JSON_HEADER = {"Content-type": "application/json"}

_SNMP_KEYS = ["v2_targets", "v3_targets", "v3_users", "communities"]

_EXIT_CODE = 0


def _set_error_exit(err, msg):
    lg.warning(msg)
    global _EXIT_CODE
    if _EXIT_CODE == 0:
        sys.stderr.write(str(err))
    # Set exit code and continue
    _EXIT_CODE = 1


def read_central_node_config_json():
    input_json = json.load(sys.stdin)
    ntp_json = None
    syslog_json_list = None
    snmp_json = None
    timezone = None
    # Convert string port numbers to integer
    if input_json.get("syslog"):
        for exporter in input_json["syslog"]["exporters"]:
            exporter["port"] = int(exporter["port"])
        syslog_json_list = []
        for syslog_json in input_json["syslog"]["exporters"]:
            syslog_json["exporter_name"] = str(uuid.uuid4())
            syslog_json_list.append(syslog_json)
    if input_json.get("snmp"):
        for v2_target in input_json["snmp"]["v2_targets"]:
            v2_target["port"] = int(v2_target["port"])
        for v3_target in input_json["snmp"]["v3_targets"]:
            v3_target["port"] = int(v3_target["port"])
        snmp_json = {
            "service_name": "snmp",
            "service_properties": input_json["snmp"]
        }
    # Create NAPI schemas
    if input_json.get("ntp"):
        ntp_json = {
            "service_name": "ntp",
            "service_properties": input_json["ntp"]
        }
    if input_json.get("timezone"):
        timezone =  input_json["timezone"]
    return ntp_json, syslog_json_list, snmp_json, timezone


def _update_time_zone(timezone):
  if timezone is not None:
    try:
        # Get existing timezone
        lg.debug("Making local API request: GET %s", _NODE_URI)
        r = requests.get(_NODE_URI)
        if r.status_code != requests.codes.ok:
            r.raise_for_status()
        current_timezone = json.loads(r.text)["timezone"]
        if timezone == current_timezone:
            lg.info("No change in timezone")
        else:
            lg.info("Updating time zone configuration")
            timezone_req = json.dumps({"timezone": timezone})
            lg.debug("Making local API request: PUT %s %s",
                      _NODE_URI, timezone_req)
            r = requests.put(_NODE_URI, data=timezone_req, headers=_JSON_HEADER)
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
    except requests.exceptions.HTTPError as err:
        _set_error_exit(err, "Failed to update time zone, response: %s, "
            "status: %s, err: %s" % (r.text, str(r.status_code), str(err)))
    except Exception:
        lg.warning("Failed to update time zone: %s", traceback.format_exc())


def _update_ntp(ntp_json):
  if ntp_json is not None:
    try:
        # Get existing NTP servers
        lg.debug("Making local API request: GET %s", _NTP_SERVICE_URI)
        r = requests.get(_NTP_SERVICE_URI)
        if r.status_code != requests.codes.ok:
            r.raise_for_status()
        current_props = json.loads(r.text)["service_properties"]
        current_props.pop("start_on_boot")
        if ntp_json["service_properties"] == current_props:
            lg.info("No change in NTP configuration")
        else:
            lg.info("Updating NTP configuration")
            ntp_req = json.dumps(ntp_json)
            lg.debug("Making local API request: PUT %s %s",
                      _NTP_SERVICE_URI, ntp_req)
            r = requests.put(_NTP_SERVICE_URI, data=ntp_req)
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
    except requests.exceptions.HTTPError as err:
        _set_error_exit(err, "Failed to update NTP service, response: %s, "
            "status: %s, err: %s" % (r.text, str(r.status_code), str(err)))
    except Exception:
        lg.warning("Failed to update NTP service: %s", traceback.format_exc())


def _update_syslog(syslog_json_list):
  if syslog_json_list is not None:
    try:
        # Get all existing syslog exporters
        lg.debug("Making local API request: GET %s", _SYSLOG_EXPORTERS_URI)
        try:
            r = requests.get(_SYSLOG_EXPORTERS_URI)
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
        except requests.exceptions.HTTPError as err:
            _set_error_exit(err, "Failed to get syslog exporter "
                "configurations, response: %s, status: %s, err: %s" % (r.text,
                str(r.status_code), str(err)))
            return
        exporters_to_delete = {}
        for exporter in json.loads(r.text)["results"]:
            props = (
                exporter["server"],
                exporter["port"],
                exporter["protocol"],
                exporter["level"]
            )
            exporters_to_delete[props] = exporter["exporter_name"]
        for syslog_json in syslog_json_list:
            props = (
                syslog_json["server"],
                syslog_json["port"],
                syslog_json["protocol"],
                syslog_json["level"]
            )
            # If an exporter already exists, don't delete it
            if props in exporters_to_delete:
                exporters_to_delete.pop(props)
                syslog_json["ignore"] = True
        # Update syslog json list
        syslog_json_list = [x for x in syslog_json_list if "ignore" not in x]
        if not exporters_to_delete and not syslog_json_list:
            lg.info("No change in syslog configuration")
        else:
            lg.info("Updating Syslog configuration")
            # Delete old syslog exporters
            for name in list(exporters_to_delete.values()):
                lg.debug("Making local API request: DELETE %s",
                         _SYSLOG_EXPORTERS_URI + "/" + name)
                try:
                    r = requests.delete(_SYSLOG_EXPORTERS_URI + "/" + name)
                    if r.status_code != requests.codes.ok:
                        r.raise_for_status()
                except requests.exceptions.HTTPError as err:
                    _set_error_exit(err, "Failed to delete syslog exporter %s, "
                        "response: %s, status: %s, err: %s" % (name, r.text,
                        str(r.status_code), str(err)))
            # Add new syslog exporters
            for syslog_json in syslog_json_list:
                syslog_req = json.dumps(syslog_json)
                lg.debug("Making local API request: POST %s %s",
                         _SYSLOG_EXPORTERS_URI, syslog_req)
                try:
                    r = requests.post(_SYSLOG_EXPORTERS_URI, data=syslog_req,
                                      headers=_JSON_HEADER)
                    if r.status_code != requests.codes.ok:
                        r.raise_for_status()
                except requests.exceptions.HTTPError as err:
                    exporter = {k:v for k, v in list(syslog_json.items())
                        if k in ["server", "port", "protocol", "level"]}
                    _set_error_exit(err, "Failed to add syslog exporter %s, "
                        "response: %s, status: %s, err: %s" % (str(exporter),
                        r.text, str(r.status_code), str(err)))
    except Exception:
        lg.warning("Failed to update syslog service: %s",
                   traceback.format_exc())


def _update_snmp(snmp_json):
  if snmp_json is not None:
    try:
        snmp_config_change = True
        # If v3_users is empty, check if there is any configuration change
        if not snmp_json["service_properties"]["v3_users"]:
            # Get existing SNMP configuration
            lg.debug("Making local API request: GET %s?show_sensitive_data=true",
                      _SNMP_SERVICE_URI)
            r = requests.get(_SNMP_SERVICE_URI + "?show_sensitive_data=true")
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
            current_props = {k:v for k, v in
                list(json.loads(r.text)["service_properties"].items())
                if k in _SNMP_KEYS}
            target_props = {k:v for k, v in
                list(snmp_json["service_properties"].items())
                if k in _SNMP_KEYS}
            if target_props == current_props:
                snmp_config_change = False
                lg.info("No change in SNMP configuration")
        if snmp_config_change:
            lg.info("Updating SNMP configuration")
            snmp_req = json.dumps(snmp_json)
            lg.debug("Making local API request: PUT %s %s",
                     _SNMP_SERVICE_URI, snmp_req)
            r = requests.put(_SNMP_SERVICE_URI, data=snmp_req,
                             headers=_JSON_HEADER)
            if r.status_code != requests.codes.ok:
                r.raise_for_status()
    except requests.exceptions.HTTPError as err:
        _set_error_exit(err, "Failed to update SNMP service, response: %s, "
            "status: %s, err: %s" % (r.text, str(r.status_code), str(err)))
    except Exception:
        lg.warning("Failed to update SNMP service: %s", traceback.format_exc())


def _update_snmp_on_kvm(snmp_json):
    lg.info("Updating SNMP configuration")
    parsed_snmp_configs = parse_snmp_configs(get_snmp_configs())
    res, http_err, err, arg1 =\
        validate_snmp_config(
            snmp_json["service_properties"], parsed_snmp_configs)
    if not res:
        _set_error_exit(err, "Failed to validate SNMP configuration")
    res, http_err, err =\
        update_snmp_config(snmp_json["service_properties"], parsed_snmp_configs)
    if not res:
        _set_error_exit(err, "Failed to update SNMP configuration")


if __name__ == "__main__":
    nsx_logging.basicConfig(syslog=True, subcomp="central_node_config_update")
    lg = nsx_logging.getLogger()
    lg.setLevel(nsx_logging.INFO)
    # Suppress INFO logs from requests
    nsx_logging.getLogger("requests").setLevel(nsx_logging.WARNING)

    # Read configuration JSON from stdin
    ntp_json, syslog_json_list, snmp_json, timezone =\
        read_central_node_config_json()

    # Update SNMP locally on KVM hosts
    if _NODE_TYPE in [NODE_TYPE_HOST_CENTOSKVM, NODE_TYPE_HOST_RHELKVM,
            NODE_TYPE_HOST_SLESKVM, NODE_TYPE_HOST_UBUNTUKVM]:
        lg.info("Ignoring timezone, NTP, and Syslog configurations on KVM "
                "hosts, updating SNMP configurations")
        _update_snmp_on_kvm(snmp_json)
    # Directly call NAPI if on appliance nodes
    elif is_appliance():
        import gevent
        import requests
        greenlets = []
        # Time zone
        greenlets.append(gevent.spawn(_update_time_zone, timezone))
        # NTP
        greenlets.append(gevent.spawn(_update_ntp, ntp_json))
        # Syslog
        greenlets.append(gevent.spawn(_update_syslog, syslog_json_list))
        # SNMP
        greenlets.append(gevent.spawn(_update_snmp, snmp_json))
        # Wait for all greenlets to finish
        gevent.joinall(greenlets)
    else:
        # Do nothing if not KVM or appliance
        pass

    # Return 0 on success and 1 on error
    if _EXIT_CODE == 0:
        lg.info("Central node config update completed successfully")
    sys.exit(_EXIT_CODE)
