# Copyright (C) 2017 VMware, Inc. All rights reserved.

import logging
from vmware.nsx.rpc import nsx_rpc_pb2

logger = logging.getLogger("nsxrpc")


class AppStatus(object):
    """ Status that NSX RPC users can pass to the library to signal
    their application-level status of the call.

    AppStatus has a subset of CallStatus codes because not all codes
    can be generated by the application. For example, COMMUNICATION_ERROR
    can only be set by the library.
    """
    # Status codes are defined and explained in nsx_rpc.proto.
    SUCCESS = nsx_rpc_pb2.RpcStatusMsg.SUCCESS
    UNKNOWN = nsx_rpc_pb2.RpcStatusMsg.UNKNOWN
    INVALID_ARGUMENT = nsx_rpc_pb2.RpcStatusMsg.INVALID_ARGUMENT
    NOT_FOUND = nsx_rpc_pb2.RpcStatusMsg.NOT_FOUND
    ALREADY_EXISTS = nsx_rpc_pb2.RpcStatusMsg.ALREADY_EXISTS
    PERMISSION_DENIED = nsx_rpc_pb2.RpcStatusMsg.PERMISSION_DENIED
    UNAUTHENTICATED = nsx_rpc_pb2.RpcStatusMsg.UNAUTHENTICATED
    RESOURCE_EXHAUSTED = nsx_rpc_pb2.RpcStatusMsg.RESOURCE_EXHAUSTED
    FAILED_PRECONDITION = nsx_rpc_pb2.RpcStatusMsg.FAILED_PRECONDITION
    ABORTED = nsx_rpc_pb2.RpcStatusMsg.ABORTED
    OUT_OF_RANGE = nsx_rpc_pb2.RpcStatusMsg.OUT_OF_RANGE
    UNIMPLEMENTED = nsx_rpc_pb2.RpcStatusMsg.UNIMPLEMENTED
    INTERNAL = nsx_rpc_pb2.RpcStatusMsg.INTERNAL
    UNAVAILABLE = nsx_rpc_pb2.RpcStatusMsg.UNAVAILABLE
    DATA_LOSS = nsx_rpc_pb2.RpcStatusMsg.DATA_LOSS

    CODE_NAMES = {
        SUCCESS: "SUCCESS",
        UNKNOWN: "UNKNOWN",
        INVALID_ARGUMENT: "INVALID_ARGUMENT",
        NOT_FOUND: "NOT_FOUND",
        ALREADY_EXISTS: "ALREADY_EXISTS",
        PERMISSION_DENIED: "PERMISSION_DENIED",
        UNAUTHENTICATED: "UNAUTHENTICATED",
        RESOURCE_EXHAUSTED: "RESOURCE_EXHAUSTED",
        FAILED_PRECONDITION: "FAILED_PRECONDITION",
        ABORTED: "ABORTED",
        OUT_OF_RANGE: "OUT_OF_RANGE",
        UNIMPLEMENTED: "UNIMPLEMENTED",
        INTERNAL: "INTERNAL",
        UNAVAILABLE: "UNAVAILABLE",
        DATA_LOSS: "DATA_LOSS"}

    def __init__(self, code, message):
        assert code in AppStatus.CODE_NAMES, (
            "AppStatus code must be one of defined values")
        self.code = code
        self.message = message

    @property
    def name(self):
        return AppStatus.CODE_NAMES[self.code]

    def __str__(self):
        if self.message:
            return "AppStatus(code=%s, message=\"%s\")" % (self.name, self.message)
        else:
            return "AppStatus(code=%s)" % self.name


class CallStatus(object):
    """ CallStatus is constructed by the NSX RPC library and passed to
    the user code to communicate call status. It must never be constructed
    by the user code. Users should use AppStatus instead.
    """

    # Codes inherited from AppStatus that can come over the wire
    SUCCESS = AppStatus.SUCCESS
    UNKNOWN = AppStatus.UNKNOWN
    INVALID_ARGUMENT = AppStatus.INVALID_ARGUMENT
    NOT_FOUND = AppStatus.NOT_FOUND
    ALREADY_EXISTS = AppStatus.ALREADY_EXISTS
    PERMISSION_DENIED = AppStatus.PERMISSION_DENIED
    UNAUTHENTICATED = AppStatus.UNAUTHENTICATED
    RESOURCE_EXHAUSTED = AppStatus.RESOURCE_EXHAUSTED
    FAILED_PRECONDITION = AppStatus.FAILED_PRECONDITION
    ABORTED = AppStatus.ABORTED
    OUT_OF_RANGE = AppStatus.OUT_OF_RANGE
    UNIMPLEMENTED = AppStatus.UNIMPLEMENTED
    INTERNAL = AppStatus.INTERNAL
    UNAVAILABLE = AppStatus.UNAVAILABLE
    DATA_LOSS = AppStatus.DATA_LOSS

    # Codes that are set by the library locally
    # Call was canceled (either by client or server).
    CANCELED = 51

    # Call timed out.
    TIMEOUT = 52

    # Local channel or whole connection is going down.
    COMMUNICATION_ERROR = 53

    # Cannot accept this message since send buffer is full.
    SEND_BUFFER_FULL = 54

    CODE_NAMES = {
        SUCCESS: "SUCCESS",
        UNKNOWN: "UNKNOWN",
        INVALID_ARGUMENT: "INVALID_ARGUMENT",
        NOT_FOUND: "NOT_FOUND",
        ALREADY_EXISTS: "ALREADY_EXISTS",
        PERMISSION_DENIED: "PERMISSION_DENIED",
        UNAUTHENTICATED: "UNAUTHENTICATED",
        RESOURCE_EXHAUSTED: "RESOURCE_EXHAUSTED",
        FAILED_PRECONDITION: "FAILED_PRECONDITION",
        ABORTED: "ABORTED",
        OUT_OF_RANGE: "OUT_OF_RANGE",
        UNIMPLEMENTED: "UNIMPLEMENTED",
        INTERNAL: "INTERNAL",
        UNAVAILABLE: "UNAVAILABLE",
        DATA_LOSS: "DATA_LOSS",
        CANCELED: "CANCELED",
        TIMEOUT: "TIMEOUT",
        COMMUNICATION_ERROR: "COMMUNICATION_ERROR",
        SEND_BUFFER_FULL: "SEND_BUFFER_FULL"}

    def __init__(self, code, message):
        assert code in CallStatus.CODE_NAMES, (
            "CallStatus code must be one of defined values")
        self.code = code
        self.message = message

    def is_success(self):
        return self.code == CallStatus.SUCCESS

    @property
    def name(self):
        return CallStatus.CODE_NAMES[self.code]

    def __str__(self):
        if self.message:
            return "CallStatus(code=%s, message=\"%s\")" % (self.name, self.message)
        else:
            return "CallStatus(code=%s)" % self.name


class CloseReason(object):
    """ This class lists reasons for which a channel/connection/socket
    could be closed.
    """

    # No specific reason provided
    UNKNOWN_REASON = 0

    # The network encountered an error at the local endpoint (e.g. the
    # connection was reset)
    NETWORK_ERROR = 1

    # We received an unexpected message we cannot recover from, or
    # another fatal error occurred at the RPC protocol level
    PROTOCOL_ERROR = 2

    # The remote endpoint closed the connection
    CLOSED_BY_REMOTE = 3

    # The object has been destroyed locally
    LOCAL_CLOSE = 4

    # Rpc/Net transport is being shut down
    SHUTDOWN = 5

    # Keepalive expired
    KEEPALIVE_EXPIRED = 6

    CODE_NAMES = {
        UNKNOWN_REASON: "UNKNOWN_REASON",
        NETWORK_ERROR: "NETWORK_ERROR",
        PROTOCOL_ERROR: "PROTOCOL_ERROR",
        CLOSED_BY_REMOTE: "CLOSED_BY_REMOTE",
        LOCAL_CLOSE: "LOCAL_CLOSE",
        SHUTDOWN: "SHUTDOWN",
        KEEPALIVE_EXPIRED: "KEEPALIVE_EXPIRED"
    }

    def __init__(self, code):
        assert code in CloseReason.CODE_NAMES, (
            "CloseReason code must be one of defined values")
        self.code = code

    @property
    def name(self):
        return CloseReason.CODE_NAMES[self.code]

    def __str__(self):
        return "CloseReason=%s" % self.name
