# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import platform
import traceback
import time
import uuid

import yaml
import sys
import nsx_logging
import subprocess
from sha.constants import CONDITION_WAVEFRONT, CONDITION_TRACE, CONDITION_TSDB, REACHED, UNREACHED_MSG, \
                          FORKSERVER_STATS, FORKSERVER_MONITOR_STATS, FORKSERVER_PROCESS_STATS, \
                          FORKSERVER_EXECUTOR_STATS, PRECONDITION_KEY
from sha.preconfig import get_sub_logger

SHA_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
LOG_LEVEL = nsx_logging.INFO
SYSTEM = platform.system()

PLATFORM_ESX = "VMkernel"
PLATFORM_LINUX = "Linux"
PLATFORM_WINDOWS = "Windows"

DIST_ESX = "nsx-esx"
DIST_KVM = "nsx-kvm"
DIST_BMS = "nsx-bms"
DIST_EDGE = "nsx-edge"
DIST_UA = "nsx-manager nsx-policy-manager nsx-controller"
DIST_GM = "global-manager"
DIST_PUBLIC_CLOUD_GATEWAY = "nsx-public-cloud-gateway"
DIST_AUTONOMOUS_EDGE = "nsx-autonomous-edge"

VERSION_ESX_65 ="6.5"
VERSION_ESX_67 ="6.7"
VERSION_ESX_70 ="7.0"

LINUX_ID_CENTOS = "centos"
LINUX_ID_RHEL = "rhel"
LINUX_ID_UBUNTU = "ubuntu"
LINUX_ID_OEL = "ol"

# some of config not used, just keep them here for later use
RHEL_VERSION_ID_76 = "7.6"
RHEL_VERSION_ID_77 = "7.7"
RHEL_VERSION_ID_78 = "7.8"
RHEL_VERSION_ID_79 = "7.9"
RHEL_VERSION_ID_82 = "8.2"
RHEL_VERSION_ID_80 = "8.0"
CENTOS_VERSION_ID_7 = "7"  # 7.6 and 7.7, 7.8
CENTOS_VERSION_ID_8 = "8"  # 8.0, 8.2
OEL_VERSION_ID_76 = "7.6"
OEL_VERSION_ID_77 = "7.7"

PLUGIN_TYPE_LISTEN = "listen"

# nsx logging has been configured in preconfig.py
_logger = get_sub_logger(__name__, s2comp="config")
_logger.info('config initializing')
_logger.info('system is {}'.format(SYSTEM))


for path in sys.path:
    _logger.info('path in sys.path: %s' % path)


# get a specific logger
def get_logger(name):
    return nsx_logging.getLogger(name)


# set memory usage soft limit
# soft, hard = resource.getrlimit(resource.RLIMIT_AS)
# resource.setrlimit(resource.RLIMIT_AS, (90 * 1024 ** 2, hard))

# retrieving HOST UUID
if platform.system() == PLATFORM_WINDOWS:
    # resolving NSX installation dir
    import winreg

    nsx_reg_path = r'SOFTWARE\VMWARE, Inc.\NSX'
    hKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, nsx_reg_path)
    nsx_path, regtype = winreg.QueryValueEx(hKey, 'InstallDir')
    if regtype != winreg.REG_SZ:
        raise TypeError(
            'NSX installation dir registry is not a string, current value: {}'.format(nsx_path))
    NSX_DIR = nsx_path
    ISSUE_PATH = os.path.join(NSX_DIR, 'nsx_issue')
    HOST_CONFIG_PATH = os.path.join(NSX_DIR, 'host-cfg.xml')
else:
    ISSUE_PATH = r'/etc/nsx_issue'
    HOST_CONFIG_PATH = r'/etc/vmware/nsx/host-cfg.xml'

BUILD_VERSION = None
# get the node type (nsx-esx, nsx-kvm, nsx-bms, nsx-edge)
with open(ISSUE_PATH) as f:
    for line in f.readlines():
        if 'version' in line and not 'build-version' in line:
            VERSION = line.split(':')[-1].strip()
            continue
        if 'node-type' in line:
            NODE_TYPE = line.split(':')[-1].strip()
        if 'build-version' in line:
            BUILD_VERSION = line.split(':')[-1].strip()
        if VERSION and NODE_TYPE and (NODE_TYPE != DIST_ESX or BUILD_VERSION):
            break
    else:
        _logger.error('can not determine which platform / version nsx sha is running on, exit now')
        raise Exception('can not determine which platform / version nsx sha is running on, exit now')

_logger.info("Got version {}, node-type: {}".format(VERSION, NODE_TYPE))
if NODE_TYPE == DIST_ESX:
    _logger.info("Got ESX build version {}.".format(BUILD_VERSION))

# begin - check LINUX_ID and LINUX_VERSION_ID
LINUX_ID = None
LINUX_VERSION_ID = None
if SYSTEM == PLATFORM_LINUX:
    OS_RELEASE_PATH = "/etc/os-release"
    with open(OS_RELEASE_PATH) as f:
        for line in f.readlines():
            if line.startswith('ID='):
                LINUX_ID = line[3:].strip()
                if LINUX_ID.startswith('"') and LINUX_ID.endswith('"'):
                    LINUX_ID = LINUX_ID[1:-1]
            if line.startswith('VERSION_ID='):
                LINUX_VERSION_ID = line[11:].strip()
                if LINUX_VERSION_ID.startswith('"') and LINUX_VERSION_ID.endswith('"'):
                    LINUX_VERSION_ID = LINUX_VERSION_ID[1:-1]

    _logger.info('got LINUX_ID:[{}], LINUX_VERSION_ID:[{}]'.format(LINUX_ID, LINUX_VERSION_ID))
    if not LINUX_ID or not LINUX_VERSION_ID:
        err_msg = 'can not determine Linux ID or version ID, exit now'
        _logger.error(err_msg)
        raise Exception(err_msg)
# end - check LINUX_ID and LINUX_VERSION_ID

# must contain all these 3 items: nsx-manager nsx-policy-manager nsx-controller
def _is_ua_type(type):
    ua_types = {'nsx-manager', 'nsx-policy-manager', 'nsx-controller'}
    ts = type.split(' ')
    if len(ts) != len(ua_types):
        return False
    for t in ts:
        if t not in ua_types:
            return False
        ua_types.remove(t)
    return True

def _is_gm_type(type):
    return type == DIST_GM

if _is_ua_type(NODE_TYPE) or _is_gm_type(NODE_TYPE):
    HOST_CONFIG_PATH = r'/etc/vmware/nsx-appl-proxy/appl-proxy-public-cfg.xml'
    if _is_ua_type(NODE_TYPE):
       NODE_TYPE = DIST_UA
       _logger.info("It's an UA node.")
    else:
       _logger.info("It's an Global Manager node.")
       NODE_TYPE = DIST_GM


def is_ua():
    return NODE_TYPE == DIST_UA

def is_gm():
    return NODE_TYPE == DIST_GM

def node_type_is(*dist_types):
    return NODE_TYPE in dist_types


def is_edge():
    return NODE_TYPE == DIST_EDGE

def is_linux_platform():
    return SYSTEM == PLATFORM_LINUX

# waiting to retrieve host id from host-cfg.xml
should_sleep = False
while True:
    HOST_ID = None
    # noinspection PyBroadException
    try:
        import xml.etree.ElementTree
        HOST_ID = xml.etree.ElementTree.parse(HOST_CONFIG_PATH).find('uuid').text
        uuid.UUID(str(HOST_ID))
        _logger.info('Host id is {}.'.format(HOST_ID))
        break
    except Exception as e:
        should_sleep = True
        _logger.warning("Could not read a valid uuid from {}, "
                     "current value:{}, error: {}, retry in 10s".format(HOST_CONFIG_PATH, HOST_ID, e))
        time.sleep(10)

_host_id = int(HOST_ID.replace('-', ''), 16)
HOST_ID_LOW = (_host_id >> 64) & 0xffffffffffffffff
HOST_ID_HIGH = _host_id & 0xffffffffffffffff

if should_sleep:
    _logger.info("It's first time start, should wait 2 minutes to avoid race in MPA client.")
    time.sleep(120)

# get hostname
try:
    HOST_NAME = subprocess.check_output(
        "hostname",
        stderr=subprocess.PIPE,
        shell=True
    ).decode('utf-8').splitlines()[0].strip()
except Exception as e:
    _logger.info('could not get hostname, error: {}'.format(e))
    HOST_NAME = HOST_ID

_logger.info('host name is {}'.format(HOST_NAME))

# get TIME_THRESHOLD for utils run_command
TIME_THRESHOLD = 180
TIME_THRESHOLD_BOTTOM = 60
if SYSTEM == PLATFORM_ESX:
    try:
        sha_config_path = os.path.join(SHA_ROOT_DIR, 'sha_config.yml')
        with open(sha_config_path, 'r') as f:
            sha_config = yaml.load(f.read(), Loader=yaml.Loader)
            time_threshold = int(sha_config.get('TIME_THRESHOLD', None))
            if time_threshold > TIME_THRESHOLD_BOTTOM:
                TIME_THRESHOLD = time_threshold
                _logger.info('Got TIME_THRESHOLD from config:{}'.format(time_threshold))
            else:
                _logger.info('Invalid TIME_THRESHOLD from config: {}, '
                             'it should be a positive integer > {}.'.format(time_threshold, TIME_THRESHOLD_BOTTOM))
    except Exception as e:
        trace = traceback.format_exc()
        _logger.error('Cannot get a valid TIME_THRESHOLD from sha_config.yml,'
                      ' use the default {} instead. exception:{}, trace:{}'.format(TIME_THRESHOLD, e, trace))
_logger.info('Use TIME_THRESHOLD:{}'.format(TIME_THRESHOLD))

# get dump file path
if SYSTEM == PLATFORM_ESX:
    DUMP_FILE_PATH = r'/var/run/log'
else:
    if is_ua() or is_edge():
        DUMP_FILE_PATH = r'/var/log/vmware'
    else:
        import tempfile
        DUMP_FILE_PATH = tempfile.gettempdir()

# get workflow trace config
WORKFLOW_TRACE_FILENAME = "nsx-workflow-span.scap"
WORKFLOW_TRACE_FILE_MAX_BYTES = 1048576 * 10
WORKFLOW_TRACE_FILE_MAX_BYTES_MINI = 1048576
WORKFLOW_TRACE_FILE_MAX_BYTES_TOP = 1048576 * 20
WORKFLOW_TRACE_FILE_MAX_BYTES_TOP_MINI = 1048576 * 10
WORKFLOW_TRACE_FILE_BACKUP_COUNT = 10
WORKFLOW_TRACE_FILE_BACKUP_COUNT_TOP = 20
WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE = 500
WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MIN = 100
WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MAX = 1000

try:
    sha_config_path = os.path.join(SHA_ROOT_DIR, 'sha_config.yml')
    with open(sha_config_path, 'r') as f:
        sha_config = yaml.load(f.read(), Loader=yaml.Loader)
        filename = sha_config.get('WORKFLOW_TRACE_FILENAME', None)
        if filename is not None:
            if filename:
                WORKFLOW_TRACE_FILENAME = filename
                _logger.info('Got WORKFLOW_TRACE_FILENAME from config:{}'.format(filename))
            else:
                _logger.info('Invalid WORKFLOW_TRACE_FILENAME: {}, '
                             'it should be an un-empty string.'.format(filename))

        # Todo: for mini system(such as SN), need get config of "WORKFLOW_TRACE_FILE_MAX_BYTES_MINI"
        max_bytes = sha_config.get('WORKFLOW_TRACE_FILE_MAX_BYTES', None)
        if max_bytes is not None:
            max_bytes = int(max_bytes)
            if max_bytes <= WORKFLOW_TRACE_FILE_MAX_BYTES_TOP and max_bytes >= 0 :
                WORKFLOW_TRACE_FILE_MAX_BYTES = max_bytes
                _logger.info('Got WORKFLOW_TRACE_FILE_MAX_BYTES from config:{}'.format(max_bytes))
            else:
                _logger.info('Invalid WORKFLOW_TRACE_FILE_MAX_BYTES from config: {}, '
                             'it should be 0 or a positive integer <= {}.'
                             .format(max_bytes, WORKFLOW_TRACE_FILE_MAX_BYTES_TOP))

        backup_count = sha_config.get('WORKFLOW_TRACE_FILE_BACKUP_COUNT', None)
        if backup_count is not None:
            backup_count = int(backup_count)
            if backup_count <= WORKFLOW_TRACE_FILE_BACKUP_COUNT_TOP and backup_count > 0:
                WORKFLOW_TRACE_FILE_BACKUP_COUNT = backup_count
                _logger.info('Got WORKFLOW_TRACE_FILE_BACKUP_COUNT from config:{}'.format(backup_count))
            else:
                _logger.info('Invalid WORKFLOW_TRACE_FILE_BACKUP_COUNT from config: {}, '
                             'it should be a positive integer <= {} .'
                             .format(backup_count, WORKFLOW_TRACE_FILE_BACKUP_COUNT_TOP))

        queue_size = sha_config.get('WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE', None)
        if queue_size is not None:
            queue_size = int(queue_size)
            if queue_size <= WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MAX and \
                queue_size >= WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MIN:
                WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE = queue_size
                _logger.info('Got WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE from config:{}'.format(queue_size))
            else:
                _logger.info('Invalid WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE from config: {}, '
                             'it should be %d ~ $d .'
                             .format(queue_size, WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MIN, WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE_MAX))
except Exception as e:
    trace = traceback.format_exc()
    _logger.error('Cannot get valid workflow tracing config from sha_config.yml,'
                  'exception:{}, trace:{}'.format(e, trace))

_logger.info('Use WORKFLOW_TRACE_FILENAME:{}, WORKFLOW_TRACE_FILE_MAX_BYTES:{},'
             ' WORKFLOW_TRACE_FILE_BACKUP_COUNT:{}, WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE:{}'
             .format(WORKFLOW_TRACE_FILENAME, WORKFLOW_TRACE_FILE_MAX_BYTES,
                     WORKFLOW_TRACE_FILE_BACKUP_COUNT, WORKFLOW_TRACE_FILEDUMP_QUEUE_SIZE))

ENABLED_METRICS = {metric[0]: metric[1] for metric in [
    ['nsx.hyperbus.hyperbus-connection-list', 'HyperbusConnectionList', [DIST_ESX, DIST_KVM, ], []],
    ['nsx.vmk.vmk-list', 'VMKList', [DIST_ESX, DIST_KVM, ], []],
    ['nsx.daemon.daemon-health', 'DaemonHealth', [DIST_ESX, DIST_KVM, DIST_BMS, ], []],
    ['nsx.ua.top-process-list', 'TopProcessList', [DIST_UA], []],
    ['nsx.ua.ua-node-latency-list', 'UaNodeLatencyList', [DIST_UA], []],
    ['nsx.host.host-metrics', 'HostMetrics', [DIST_ESX, ], []],
    ['nsx.vdr.vdr-stats', 'VdrStats', [DIST_ESX, ], []],
    ['nsx.edge.edge-fw-stats', 'EdgeFWStats', [DIST_EDGE, ], []],
    ['nsx.edge.edge-fw-conn', 'EdgeFWConn', [DIST_EDGE, ], []],
    ['nsx.edge.edge-ipsec-vpn-stats', 'EdgeIpsecVpnStats', [DIST_EDGE, ], []],
    ['nsx.edge.edge-status', 'EdgeStatus', [DIST_EDGE, ], []],
    ['nsx.portmirror.mirror-stack-status', 'MirrorStackStatus', [DIST_ESX, ], []],
    ['nsx.auditlog.auditlog-health-for-local-error-status', 'AuditLogHealthForLocalErrorStatus',
     [DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], []],
    ['nsx.auditlog.auditlog-health-for-remote-error-status', 'AuditLogHealthForRemoteErrorStatus',
     [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], []],
    ['nsx.communication.manager-fqdn-lookup-failure-status', 'ManagerFqdnLookupFailureStatus',
     [DIST_BMS, DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], []],
    ['nsx.communication.manager-fqdn-reverse-lookup-failure-status', 'ManagerFqdnReverseLookupFailureStatus',
     [DIST_UA, DIST_GM, ], []],
    ['nsx.edge.edge-sr-state', 'EdgeSrState', [DIST_EDGE, ], []],
    ['nsx.edge.mgmt-interface-status', 'EdgeMgmtInterface', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_AUTONOMOUS_EDGE, ], []],
    ['nsx.edge.rcu-event-status', 'EdgeRcuBlock', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_AUTONOMOUS_EDGE, ], []],
    ['nsx.common.sys-stats', 'SysStats', [DIST_EDGE, DIST_UA, ], []],
    ['nsx.edge.edge-dp-stats', 'EdgeDpStats', [DIST_EDGE, ], []],
    ['nsx.nestdb.nestdb-status', 'NestdbStatus', [DIST_ESX, DIST_KVM, DIST_BMS], [PLATFORM_ESX, PLATFORM_LINUX]],
    ['nsx.edge.edge-lr-port-stats', 'EdgeLRPortStats', [DIST_EDGE, ], []],
    ['nsx.edge.tls-stats', 'TlsStats', [DIST_EDGE, ], []],
    ['nsx.lb-status', 'LbStatus', [DIST_EDGE, DIST_ESX,  ], []],
    ['nsx.edge.idps-stats', 'IdpsStats', [DIST_EDGE, ], []],
] if NODE_TYPE in metric[2] and (not metric[3] or (SYSTEM in metric[3]))}

ENABLED_MONITORS = {monitor[0]: monitor[1] for monitor in [
    ['UplinkStatusMonitor', 'nsx.uplink.uplink-status', [DIST_ESX, DIST_KVM, DIST_BMS]],
    ['MetricsMonitor', ['nsx.metrics.nsx-cfgagent.liveness',
                        'nsx.metrics.nsx-opsagent.liveness',
                        'nsx.metrics.nsx-cfgagent.heapmemory.allocated',
                        'nsx.metrics.nsx-opsagent.heapmemory.allocated',
                        'nsx.metrics.nestdb-init.nestdb_is_started'],
     [DIST_ESX, DIST_EDGE, DIST_KVM, DIST_BMS, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['TraceMonitor', 'nsx.trace.spans', [DIST_ESX, DIST_EDGE, DIST_UA, ], [VERSION_ESX_70]],
] if NODE_TYPE in monitor[2] and (len(monitor) == 3 or (NODE_TYPE != DIST_ESX or BUILD_VERSION in monitor[3]))}

ENABLED_ACTIONS = [action[0] for action in [
    ['Hbreport', [DIST_ESX, DIST_KVM, ]],
    ['Nareport', [DIST_ESX, DIST_KVM, ]],
    ['Dhreport', [DIST_ESX, DIST_KVM, DIST_BMS, ]],
    ['Alarm', [DIST_ESX, DIST_KVM, DIST_BMS, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ]],
    ['Exportdata', [DIST_EDGE, DIST_ESX, DIST_UA, ]],
    ['Msreport', [DIST_ESX, ]],
    ['Tpreport', [DIST_UA]],
    ['UaNLReport', [DIST_UA]],
    ['Remedy', [DIST_ESX, DIST_KVM, DIST_BMS, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, ]],
] if NODE_TYPE in action[1]]

PRE_INSTALL_SCRIPTS = {os.path.join(SHA_ROOT_DIR, 'plugins', plugin[0] + '.py'): plugin[0] for plugin in [
    ['hyperbus_monitor', [DIST_ESX, DIST_KVM, ], [], [], []],
    ['node_agent_monitor', [DIST_ESX, DIST_KVM, ], [], [], []],
    ['daemon_health_monitor', [DIST_ESX, DIST_KVM, DIST_BMS, ], [], [], []],
    ['uplink_status_monitor', [DIST_ESX, DIST_KVM, DIST_BMS, ], [], [], []],
    ['host_metrics_monitor', [DIST_ESX, ], [], [], []],
    ['vdr_stats_monitor', [DIST_ESX, ], [], [], []],
    ['edge_fw_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['edge_fw_conn_monitor', [DIST_EDGE, ], [], [], []],
    ['edge_ipsec_vpn_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['edge_status_monitor', [DIST_EDGE, ], [], [], []],
    ['mirrorstack_monitor', [DIST_ESX, ], [], [], []],
    ['top_process_monitor', [DIST_UA], [], [], []],
    ['audit_log_health_for_local_monitor', [DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], [], [], []],
    ['audit_log_health_for_remote_monitor', [DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], [], [], []],
    ['manager_fqdn_lookup_failure_monitor', [DIST_BMS, DIST_ESX, DIST_KVM, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], [], [], []],
    ['manager_fqdn_reverse_lookup_failure_monitor', [DIST_UA, DIST_GM, ], [], [], []],
    ['ua_node_latency_monitor', [DIST_UA], [], [], []],
    ['edge_l3_monitor', [DIST_EDGE, ], [], [], []],
    ['edge_appliance_monitor', [DIST_EDGE, ], [], [], []],
    ['edge_mgmt_interface_monitor', [DIST_EDGE, DIST_AUTONOMOUS_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, ], [], [], []],
    ['rcu_event_monitor', [DIST_EDGE, DIST_AUTONOMOUS_EDGE, DIST_PUBLIC_CLOUD_GATEWAY, ], [], [], []],
    ['process_status_monitor', [DIST_ESX, DIST_EDGE, DIST_KVM, DIST_BMS, DIST_PUBLIC_CLOUD_GATEWAY, DIST_UA, DIST_GM, ], [], [], []],
    ['service_opsagent_monitor', [DIST_ESX, ], [], [], []],
    ['service_cfgagent_monitor', [DIST_ESX, ], [], [], []],
    ['trace_monitor', [DIST_ESX, DIST_EDGE, DIST_UA, ], [VERSION_ESX_70], [], []],
    ['sys_stats_monitor', [DIST_EDGE, DIST_UA, ], [], [], []],
    ['tsdb_monitor', [DIST_EDGE, DIST_UA, ], [], [], []],
    ['edge_dp_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['nestdb_remedy', [DIST_ESX, DIST_KVM, DIST_BMS], [], [PLATFORM_ESX, PLATFORM_LINUX], [LINUX_ID_CENTOS, LINUX_ID_RHEL, LINUX_ID_OEL]],
    ['edge_lr_port_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['tls_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['lb_status', [DIST_ESX, DIST_EDGE, ], [], [], []],
    ['idps_stats_monitor', [DIST_EDGE, ], [], [], []],
    ['wavefront_monitor', [DIST_EDGE, DIST_UA, ], [], [], []],
] if NODE_TYPE in plugin[1]
                        and (not plugin[2] or (NODE_TYPE != DIST_ESX or BUILD_VERSION in plugin[2]))
                        and (not plugin[3] or (SYSTEM in plugin[3]))
                        and (not plugin[4] or (SYSTEM != PLATFORM_LINUX or LINUX_ID in plugin[4]))}

GLOBAL_MINIMUM_CHECKING_INTERVAL = 1
# max num for custom plugin
MAX_INSTALLED_CUSTOM_PLUGIN_NUM = 10

PRE_CONDITION = {
    CONDITION_WAVEFRONT: {REACHED: False, UNREACHED_MSG: 'Wavefront collector has not been set'},
    CONDITION_TRACE: {REACHED: False, UNREACHED_MSG: 'Workflow trace has not been enabled'},
    CONDITION_TSDB: {REACHED: False, UNREACHED_MSG: 'TSDB collector has not been set'}
}

# SHA private metrics, including SHA stats and state info, not collected via ENABLED_METRICS
SHA_PRVIATE_METRICS = {
    FORKSERVER_STATS,
    FORKSERVER_MONITOR_STATS,
    FORKSERVER_PROCESS_STATS,
    FORKSERVER_EXECUTOR_STATS,
    PRECONDITION_KEY
}

# summarize all metric names from monitor
MONITORED_METRICS = set()
for metric_name in ENABLED_MONITORS.values():
    if isinstance(metric_name, list):
        MONITORED_METRICS |= set(metric_name)
    else:
        MONITORED_METRICS.add(metric_name)

# summarize all existing metric names
EXISTING_METRICS = SHA_PRVIATE_METRICS | ENABLED_METRICS.keys() | MONITORED_METRICS

# Metrics, such as the ones use NsxRpcClient, need use dedicated thread to handle
DEDICATED_THREAD_METRICS = ['nsx.daemon.daemon-health']

# available remediation CMD according to node type and platform
# cmd format: [<class_name>, <module_name>, <applicable_node_types>, <applicable_platforms>]
ENABLED_RMDY_CMD = {cmd[0]: cmd[1] for cmd in [
    ['RestartNestDb', 'restart_nestdb',
     {DIST_ESX, DIST_KVM, DIST_BMS, DIST_EDGE, DIST_PUBLIC_CLOUD_GATEWAY},  # applicable node types
     {PLATFORM_ESX, PLATFORM_LINUX}],  # applicable platforms
] if NODE_TYPE in cmd[2] and SYSTEM in cmd[3]}

#prefix list for dedicated passive-listening plugins
PUSH_METRICS_PREFIX_LIST = ['tsdb', 'wavefront']

_logger.info('config initialized')
