# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

from jaeger_client.thrift_gen.agent import Agent
from thrift.transport.TTransport import TMemoryBuffer
from thrift.protocol import TCompactProtocol, TBinaryProtocol, TJSONProtocol
from jaeger_client.thrift_gen.agent.Agent import Processor
import uuid
import struct
import sha.config
import asyncio

class Handler(Agent.Iface):
    """
    Iface instance, formatting and emitting span data
    """
    _queue = None

    def __init__(self, queue):
        super(Handler, self).__init__()
        self._queue = queue
        self._logger = sha.config.get_logger(__name__)

    def emitZipkinBatch(self, batch):
        self._logger.info("TM: zipkin is not supported")
        pass

    def emitBatch(self, batch):
        self._logger.info("TM: emit batch {}".format(batch))
        ans = []

        process = batch.process.serviceName
        for span in batch.spans:
            try:
                fs = self.formatSpan(span, process)
                ans.append(fs)
            except Exception as e:
                self._logger.warning("TM: Format span error, Exception:{}, trace:{}".format(e, trace))
        self._logger.info("TM: Span data: {}".format(ans))
        self._queue.put({"nsx.trace.spans": ans})

    def formatSpan(self, span, process):
        record = {}
        traceId = self.toUUID(span.traceIdHigh, span.traceIdLow)
        spanId = self.toUUID(0, span.spanId)
        record['name'] = span.operationName
        record['source'] = 'localhost'
        record['traceId'] = traceId
        record['spanId'] = spanId
        if (span.parentSpanId != 0):
            record['parent'] = self.toUUID(0, span.parentSpanId)

        record['cluster']= "none"
        record['shard'] = "none"
        tags = [('application', 'NSX')]
        for tag in span.tags:
            t = tag.vType
            if t == 0:
                tags.append((tag.key, tag.vStr))
            elif t == 1:
                tags.append((tag.key, str(tag.vDouble)))
            elif t == 2:
                tags.append((tag.key, 'true' if tag.vBool else 'false'))
            elif t == 3:
                tags.append((tag.key, str(tag.vLong)))
            elif t == 4:
                tags.append((tag.key, tag.vBinary))
        record['tags'] = tags
        record['timestamp'] = int(round(span.startTime / 1000))
        record['duration'] = int(round(span.duration / 1000))
        return record

    def toUUID(self, dataHigh, dataLow):
        return str(uuid.UUID(bytes = struct.pack('!qq', dataHigh, dataLow)))


class SpanProcessService:
    """
    Span processor service, handling the received span data
    """
    def __init__(self, queue):
        handler = Handler(queue)
        self.processor = Processor(handler)
        self._logger = sha.config.get_logger(__name__)
    def process(self, data):
        self._logger.info("TM: SpanProcessService.process {}".format(data))
        recvBuf = TMemoryBuffer(data)
        recvProtocol = TCompactProtocol.TCompactProtocol(recvBuf)
        exceptBuf = TMemoryBuffer()
        exceptProtocol = TJSONProtocol.TJSONProtocol(exceptBuf)
        asyncio.set_event_loop(asyncio.new_event_loop())
        self.processor.process(recvProtocol, exceptBuf)
