# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import os
import traceback

import sha.config
from sha.config import PRE_CONDITION
from sha.constants import CERTIFICATE, ENDPOINT, NSX_CLUSTER_ID, NODE_KEY, NODE_CERT


class HostConfigResolver:
    """
       Resolve CommonAgentHostConfigMsg from nestdb
            or TsdbConfigMsg from UA via MPA
    """
    _logger = sha.config.get_logger(__name__)
    _ingress_conf = {}

    @classmethod
    def _is_changed(cls, ingress_conf):
        fields =[CERTIFICATE, NSX_CLUSTER_ID, NODE_CERT, NODE_KEY]
        for key in fields:
            if cls._ingress_conf.get(key, None) != ingress_conf.get(key, None):
                return True
        if set(cls._ingress_conf[ENDPOINT]) != set(ingress_conf[ENDPOINT]):
            return True

        return False

    @classmethod
    def set(cls, profile_pb2):
        """
        :param profile_pb2: protobuf_msg
        """
        ingress_conf = {}
        try:
            if profile_pb2.HasField('ingress_certificate'):
                ingress_conf[CERTIFICATE] = profile_pb2.ingress_certificate
            ingress_conf[ENDPOINT] = profile_pb2.ingress_endpoint
            if profile_pb2.HasField('nsx_cluster_id'):
               ingress_conf[NSX_CLUSTER_ID] = profile_pb2.nsx_cluster_id

            # Handle private key and certificate of UA/GM
            if sha.config.is_ua() or  sha.config.is_gm():
                if profile_pb2.HasField('nsx_certificate'):
                    ingress_conf[NODE_CERT] = profile_pb2.nsx_certificate
                if profile_pb2.HasField('nsx_private_key'):
                    ingress_conf[NODE_KEY] = profile_pb2.nsx_private_key

            cls._logger.debug("Get Ingress certificate conf.")
            if not cls._is_changed(ingress_conf):
                cls._logger.info("No change in ingress certificate conf.")
                return

            cls._ingress_conf = ingress_conf
            # Only update TSDB certificate on supported platforms
            if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
                from sha.core.channel.provider.tsdb_provider import TsdbSender
                TsdbSender.update_cert(ingress_conf)
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.error('Failed in set profile_pb2:{}, trace:{}'.format(str(profile_pb2), trace))

    @classmethod
    def clear(cls, old_profile_pb2=None):
        """
        :param profile_pb2: old_protobuf_msg
        """
        try:
            if cls._ingress_conf:
                cls._ingress_conf = {}
                if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
                    from sha.core.channel.provider.tsdb_provider import TsdbSender
                    TsdbSender.update_cert()
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.error('Failed in clear old_profile_pb2:{}, trace:{}'.format(str(old_profile_pb2), trace))
