# The necessary boilerplate for creating a new interface is:

#   import interface
#   import some_pb2.py
#
#   class InterfaceName(interface.Interface):
#       def __init__(self):
#           interface.Interface.__init__(self, some_pb2, some_pb2.stub)

# You will also need to add an instance of the interface to the interface_list
# in the main file, and to make sure that required *_pb2.py and *.rpc.pb are
# added to the nsx-proxy Makefile.

import re
import vmware.nsx.cli.options_pb2 as cli_options

class Interface(object):
    """
    A superclass for CLI interfaces.
    """

    def __init__(self, module=None, stub=None):
        self.client = None # initiated before each command call
        self.endpoint = None # populated when initialize CliTool
        self.module = module
        self.stub = stub

        self.make_proto_fns()

    def err_check(self):
        """
        Called at the start of all generated functions. Should be manually
        called at the beginning of custom functions of Interface classes.
        """
        if self.client is None:
            return "Client not initialized in interface!\n"
        if self.module is None:
            return "Module not initialized in interface!\n"

    def make_proto_fns(self):
        """
        Generates interface functions from protobuf service methods which begin
        with 'Get'.
        """
        service_descriptor = self.stub.DESCRIPTOR
        service_name = service_descriptor.name

        for m in service_descriptor.methods:
            if re.match(r"Get.*", m.name):
                self.make_get_fn(m, service_name)

    # Limitations:
    # -- Does not support input arguments of 'fixed', 'group', or 'message' types.
    # -- Does not support multiple /actually/ optional arguments, because we
    # map arguments given in the command string to protobuf message fields in
    # order. We also cannot support multiple repeated fields, and a single
    # repeated field must always be the last field in a protobuf message. E.g.:

       # message {
       #         required string arg1;
       #         optional string arg2;
       #         }
       # works as expected.

       # message {
       #         required string arg1;
       #         repeated string arg2;
       #         }
       # works as expected.

       # message {
       #         optional string arg1;
       #         optional string arg2;
       #         repeated string arg2;
       #         }
       # works if arg1 only exists when arg2 does, and if arg3 only exists when
       # arg1 and arg2 do.

       # message {
       #         repeated string arg1;
       #         optional string arg2;
       #         }
       # does not work, because every argument will be considered part of the
       # repeated arg1 field.

    # If you need a non-supported argument ordering or type, you can create a
    # custom interface method to handle it.
    def make_get_fn(self, method, service_name):
        """
        Generates an 'get_' function for all protobuf service methods which
        begin with 'Get'. These functions are detected by the Client Engine,
        which generates corresponding CLI commands.
        """
        req_name = method.input_type.name
        fields = [f for f in method.input_type.fields]

        get_name = method.name
        fn_name = "get_" + get_name[3:]
        rsp_name = service_name + '_' + method.name

        def field_types(type_number):
            return  {
                    1: float, #double
                    2: float,
                    3: int, #int64
                    4: int, #uint64
                    5: int, #int32
                    # 6: fixed64
                    # 7: fixed32
                    8: bool,
                    9: str,
                    # 10: group
                    # 11: message
                    12: bytes,
                    13: int, #uint32
                    14: int, #enum
                    # 15: sfixed32
                    16: int, #sint64
                    17: int, #sint32
                    18: int, #sint64
                    }.get(type_number, 0)

        def fn(args):
            self.err_check()
            req = getattr(self.module, req_name)()

            # protobuf fields store 'labels' as numbers:
            # optional == 1, required == 2, repeated == 3
            for i in range(len(fields)):
                if len(args) > i:

                    # assign all non-repeated field arguments after casting to
                    # correct types
                    if fields[i].label != 3:
                        arg = field_types(fields[i].type)(args[i])
                        setattr(req, fields[i].name, arg)

                    # if we have a repeated field, cast and assign remaining
                    # arguments to it
                    else:
                        repeats = [field_types(fields[i].type)(a) for a in args[i:]]
                        getattr(req, fields[i].name).extend(repeats)
                        break

            rsp = getattr(self.client, rsp_name)(req)
            return rsp

        setattr(self, fn_name, fn)
        fn.__name__ = fn_name

        name = get_name[3:] # strip "get_"
        self.make_help_fn(method, name)

    def make_help_fn(self, method, name, info=None):
        """
        Generates a 'help' function that prints field types for the input and
        output protobuf message types, any information specified as a cli_help
        option in the protobuf method descriptor, and any additional
        information given in 'info'. The Client Engine will automatically
        generate a 'help CommandName' command that will call these generated
        functions.

        NOT meant to be called directly. To generate help functions for custom
        commands, use 'make_default_help'.
        """
        fn_name = "help_" + name

        def fn():
            if info:
                print('\n' + info)

            if method.GetOptions().HasExtension(cli_options.help):
                print('\nProtobuf Message: ' + method.name + '\n' + method.GetOptions().Extensions[cli_options.help] + '\n')

            label = {
                    1: "optional",
                    2: "required",
                    3: "repeated",
                    }

            field_type = {
                    1: "double",
                    2: "float",
                    3: "int64",
                    4: "uint64",
                    5: "int32",
                    6: "fixed64",
                    7: "fixed32",
                    8: "bool",
                    9: "string",
                    10: "group",
                    11: "message",
                    12: "bytes",
                    13: "uint32",
                    14: "enum",
                    15: "sfixed32",
                    16: "sint64",
                    17: "sint32",
                    18: "sint64",
                    }

            print("Arguments:")
            for field in method.input_type.fields:
                print("  " + field.name + ":" \
                    + "\n      label: " + label.get(field.label) \
                    + "\n      type: " + field_type.get(field.type))
                if field.has_default_value:
                    print("\n      default value: " + str(field.default_value))

            print("\nReturn Types:")
            for field in method.output_type.fields:
                print("  " + field.name + ":" \
                    + "\n      label: " + label.get(field.label) \
                    + "\n      type: " + field_type.get(field.type))
                if field.has_default_value:
                    print("      default value: " + str(field.default_value))

        setattr(self, fn_name, fn)

    # for example, see: MetricTable in interfaces/proxy_metrics.py
    def make_default_help(self, proto_name, help_name, info=None):
        """
        To be called in __init__ of interface modules for custom functions
        without custom 'help' commands.

        proto_name: string name of the protobuf rpc method used by the
        interface method

        command_name: string name that will be used to call
        the help function, e.g. 'help CommandName'

        info: string of any additional information to be printed by the help
        function
        """
        for m in self.stub.DESCRIPTOR.methods:
            if re.match(m.name, proto_name):
                self.make_help_fn(m, help_name, info)
