# Copyright (C) 2016-2018 VMware, Inc. All rights reserved.

import collections
import logging
import numbers
import ctypes
import os

from vmware.nsx.rpc.status import CloseReason
from vmware.nsx.rpc.client.channel import NsxRpcChannel
from vmware.nsx.rpc.client.channel import ClientContext
from vmware.nsx.rpc.client.channel import SyncCallContext
from vmware.nsx.rpc.client.channel import AsyncCallContext
from vmware.nsx.rpc.client.transport import NsxRpcConnection

ASYNC_METHOD_POSTFIX = "_async"
DEFAULT_TIMETOUT = 10

LOGGER = logging.getLogger("nsxrpc")
# The library path may be different depeding on whether it is running on a
# building machine or on a host machine
BUILD_PATH = os.environ["NSX_SHARED_LIBS_ROOT"] + "/" \
             if "NSX_SHARED_LIBS_ROOT" in os.environ else "/"
LIBCRYPTOS = [BUILD_PATH + "opt/vmware/nsx-shared/lib64/libcrypto.so", # KVM
              BUILD_PATH + "usr/lib/vmware/nsx-shared/lib64/libcrypto.so", # ESX
              BUILD_PATH + "usr/lib/x86_64-linux-gnu/libcrypto.so", # DEBIAN
              BUILD_PATH + "lib/x86_64-linux-gnu/libcrypto.so.1.0.0"] # EDGE/PCG
SYSTEM_FIPS_FILE = "/etc/vmware/system_fips"


class NsxRpcClient(object):
    """
    NsxRpcClient allows users to make RPC calls to one or more remote services.
    User must explicitly call Destroy() when the client is no longer needed.

    There are two valid syntax than can be used for NsxRpcClient objects. If
    the user wishes to only connect to one remote service, they can pass in a
    single Stub to the creator and will have no function name alterations.

    To connect to multiple services, the user can pass in a list of Stubs. This
    additionally requires the user to prefix function calls to the NsxRpcClient
    with "ServiceName_", e.g. a call to a function GetItem of the service
    MyService should written 'MyNsxRpcClient.MyService_GetItem(request)'

    When an endpoint is specified, NsxRpcClient will create a connection to it.
    By default NsxRpcClient does not attempt to reestablish this connection if
    it fails during the initial Connect() call or later during the connection
    lifecycle. This can be changed by specifying a RetryPolicy with the number
    of reconnection attempts and the delay between them.

    Whether NsxRpcClient enters FIPS mode by default is controlled by the
    presence of /etc/vmware/system_fips. If this indicator exists, NsxRpcClient
    will run in FIPS mode. During the execution of current process, whether or not
    FIPS mode is on will be consistant.

    Note: Service names using multistub syntax CANNOT contain underscores.
    """
    def __init__(self, Stub, endpoint=None, connection=None,
                 timeout=DEFAULT_TIMETOUT, sec_ctx=None,
                 retry_policy=None, destination=None):
        LOGGER.debug(
            "NsxRpcClient::Constructor (Stub:%s endpoint:%s connection:%s)",
            Stub, endpoint, connection)

        assert bool(endpoint) != bool(connection), (
            "Exactly one of endpoint or connection must be provided")

        self.connection = connection

        self.stub_info = {}
        StubInfo = collections.namedtuple('StubInfo', ['channel', 'stub'])

        self.msyntax = None
        self.own_connection = None

        # Determine whether the client should run in FIPS mode
        global SYSTEM_FIPS_FILE, LIBCRYPTOS
        self.is_FIPS_compliant = os.path.isfile(SYSTEM_FIPS_FILE)
        found_lib = False
        for lib in LIBCRYPTOS:
          if os.path.isfile(lib):
            self.FIPS_LIB = ctypes.CDLL(lib)
            found_lib = True
            break

        if not found_lib and not self.is_FIPS_compliant:
            LOGGER.debug("NsxRpcClient::__init__ OpenSSL FIPS library not found, \
                          /etc/vmware/system_fips not present. Run in non-FIPS mode.")
        elif not found_lib and self.is_FIPS_compliant:
            LOGGER.debug("NsxRpcClient::__init__ OpenSSL FIPS library not found while \
                          /etc/vmware/system_fips is present. Library paths:" + str(LIBCRYPTOS))
            raise EnvironmentError("Invalid paths of OpenSSL library")
        elif found_lib and not self.is_FIPS_compliant:
            LOGGER.debug("NsxRpcClient::__init__ OpenSSL FIPS library found, \
                          /etc/vmware/system_fips not present. Run in non-FIPS mode.")
            self.fips_off()
        else:
            LOGGER.debug("NsxRpcClient::__init__ OpenSSL FIPS library found, \
                          /etc/vmware/system_fips is present. Run in FIPS mode.")
            self.fips_on()

        if connection is None:
            LOGGER.debug("NsxRpcClient::Constructor using own connection")
            self.own_connection = NsxRpcConnection()
            self.own_connection.Connect(endpoint, sec_ctx=sec_ctx, retry_policy=retry_policy)
            self.connection = self.own_connection
        else:
            self.connection = connection

        # If we got a list then we have multiple services (or a single service
        # given to us in the new syntax). We use a bool to make syntax
        # decisions instead of the number of stubs so that users can use the
        # multistub syntax with a single stub if they pass that stub in as a
        # list.
        stub_list = Stub if isinstance(Stub, list) else [Stub]
        for st in stub_list:
            name = st.GetDescriptor().name
            if (isinstance(Stub, list)) and ('_' in name):
                raise ValueError("%s: service names cannot contain '_' when using multiple stubs" % name)
            channel = NsxRpcChannel(
                    self.connection, st.GetDescriptor().full_name, destination)
            self.connection.RegisterChannel(channel.stream_id, channel)
            self.stub_info[name] = StubInfo(channel, st(channel))

        if isinstance(Stub, list):
            self.msyntax = True
        else:
            self.msyntax = False

            # support for old tests/use cases
            self.channel = channel
            self.stub = Stub(channel)

        self.timeout = timeout

    def fips_on(self):
        """Call C API to enable OpenSSL FIPS mode. User of RpcClient should not call
        this function directly."""
        try:
            LOGGER.debug("NsxRpcClient::fips_on Turning on FIPS mode")
            self.FIPS_LIB.FIPS_mode_set(1)
            if 0 == self.FIPS_LIB.FIPS_mode():
                LOGGER.debug("NsxRpcClient::fips_on Failed to turn on FIPS mode")
                self.is_FIPS_compliant = False
                raise RuntimeError("Failed to enable FIPS mode")
            else:
                LOGGER.debug("NsxRpcClient::fips_on Successfully turned on FIPS mode")
                self.is_FIPS_compliant = True
        except Exception as e:
            LOGGER.debug("NsxRpcClient::fips_on FIPS_mode_set() raised an exception:" + str(e))
            self.is_FIPS_compliant = False
            raise RuntimeError("Unable to call APIs from the FIPS module")

    def fips_off(self):
        """Call C API to disable OpenSSL FIPS mode. User of RpcClient should not call
        this function directly."""
        try:
            LOGGER.debug("NsxRpcClient::fips_off Turning off FIPS mode")
            self.FIPS_LIB.FIPS_mode_set(0)
            LOGGER.debug("NsxRpcClient::fips_off Turned off FIPS mode")
        except Exception as e:
            LOGGER.debug("NsxRpcClient::fips_off FIPS_mode_set() raised an exception:" + str(e))
            raise RuntimeError("Unable to call APIs from the FIPS module")
        finally:
            self.is_FIPS_compliant = False

    def fips_status(self):
        """Return True if this client is in FIPS mode."""
        return self.is_FIPS_compliant

    def __getattr__(self, name):
        LOGGER.debug("NsxRpcClient::__getattr__ (name:%s)", name)

        sync = True
        if name.endswith(ASYNC_METHOD_POSTFIX):
            name = name[:-len(ASYNC_METHOD_POSTFIX)]
            sync = False

        # check if we have multiple stubs/are using multistub syntax
        if self.msyntax:
            service = None
            try:
                service, method = name.split('_', 1)
            except:
                print(("Expected function name in format {service name}_{method name}, got {}".format(name)))

            if (service is None) or (service == ''):
                raise AttributeError("Stub for service '%s' does not exist" % (service))

        else:
            service = next(iter(self.stub_info))
            method = name # single service/old syntax case

        method_descriptor = self.stub_info[service].stub.GetDescriptor().FindMethodByName(method)

        if method_descriptor is None:
            raise AttributeError("'%s' object has no attribute '%s'" % (
                self.stub_info[service].stub.__class__.__name__, name))

        if sync:
            return self.SyncMethodProxy(method_descriptor, service)
        else:
            return self.AsyncMethodProxy(method_descriptor, service)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """If the block is exited because of an exception, we let it
        bubble up normally (by returning False). We simply make sure
        that the client is destroyed.
        """
        self.Destroy()
        return False

    def SyncMethodProxy(self, method_descriptor, service=None):

        if service is None:
            # if no service specified, assume only one service exists
            service = next(iter(self.stub_info))

        method = getattr(self.stub_info[service].stub, method_descriptor.name)

        def call(request, timeout=None, client_context=None):

            assert (request is None or
                    isinstance(request,
                               self.stub_info[service].stub.GetRequestClass(method_descriptor))), (
                   "Synchronous RPC call: expected request type '%s' given '%s'"
                       % (self.stub_info[service].stub.GetRequestClass(method_descriptor),   # noqa: E127
                          request.__class__))

            timeout = timeout or self.timeout
            # We allow timeout to be None here to signify that user wants no
            # timeout. This is not super dangerous because user needs to
            # explicitly set timeout=None when creating the client for this to
            # happen. Thus, having timeout=None can't be accidental.

            client_context = client_context or ClientContext()

            # callback argument is used to pass call context
            return method(rpc_controller=None,
                          request=request,
                          callback=SyncCallContext(timeout, client_context))

        return call

    def AsyncMethodProxy(self, method_descriptor, service=None):

        if service is None:
            service = next(iter(self.stub_info))

        method = getattr(self.stub_info[service].stub, method_descriptor.name)

        def call(request, callback, client_context=None):
            assert (request is None or
                    isinstance(request,
                               self.stub_info[service].stub.GetRequestClass(method_descriptor))), (
                   "Asynchronous RPC call: expected request type '%s' given '%s'"
                       % (self.stub_info[service].stub.GetRequestClass(method_descriptor),  # noqa: E131
                          request.__class__))

            assert isinstance(callback, collections.Callable), (
                "Asynchronous RPC call: expected callback, given '%s'"
                % callback)

            client_context = client_context or ClientContext()

            # callback argument is used to pass call context
            return method(rpc_controller=None,
                          request=request,
                          callback=AsyncCallContext(callback, client_context))

        return call

    def Destroy(self):
        """ Closes all open channels. """
        LOGGER.debug("NsxRpcClient::Destroy")
        for st in self.stub_info:
            self.stub_info[st].channel.Close(CloseReason(CloseReason.LOCAL_CLOSE))
        if self.own_connection:
            LOGGER.debug("NsxRpcClient::Destroy closing own connection")
            self.own_connection.Close(CloseReason(CloseReason.LOCAL_CLOSE))

    def CreateClientContext(self):
        return ClientContext()
